// fileio.c
// functions for controlling Disk IO
//
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dos.h>
#include <conio.h>
#include <ctype.h>
#include <time.h>
#include <fcntl.h>
#include <io.h>
#include <share.h>
#include <sys\types.h>
#include <sys\stat.h>

#include <GRconst.h>
#include <GRports.h>
#include <GRextrn.h>
#include <menuet.h>

#include "navdefs.h"
#include "navtypes.h"
#include "nav_util.h"
#include "filehndl.h"
#include "fileio.h"
#include "serial.h"
#include "mnutils.h"
#include "graph.h"
#include "dispasci.h"
#include "printer.h"
#include "nav.h"


/* Global data */
static FILE * BinNavLogFP  = (FILE *) NULL;              // binary nav log file
static FILE * AsciiLogFP   = (FILE *) NULL;              // Ascii log file

static char * NULL_SETUP_MSG  = "Null Configuration";
static char * NULL_COM_MSG = "Null COM pointer";
static char * NULL_NAV_MSG = "Null NAV pointer";

extern t_com com_struct[];
extern t_nav navSensors[];
extern t_nav * curNav;
extern int depthAlarmFlag;

char * sensorAbbrevs[] = 
	{
  "Mx 1107",
  "Asht GPS",
  "DDMU 540",
  "NS 7000 ",
  "4000AX",
  "TrkPt",
  "Bridge Terminal",
  "DataSonics Chirp",
  "KVH Compass",
  "DelNorte to URI",
  "Trackpoint to URI",
  "BUSTER_GYRO",
  "TRIMBLE_NAVTRAK",
  "TRIMBLE_10X"
	};



/*----------------------------------------------------------------------------*/
// record the nav data for the passed sensor
/*----------------------------------------------------------------------------*/
int recordNav(t_com * p)
{
static double lastLogTime[MAX_COMPORTS];
int k;
char buf[128];
static time_t lastWarningTime = 0L;
unsigned long bytesFree;
t_setup * config = QuerySetup();
unsigned int drive;
time_t now = time(NULL);

if (config->BinLoggingEnabled)
	{
	// is the log file open
	if (BinNavLogFP == (FILE *) NULL)
		{
		return BAD_FILE_POINTER;
		}

	// is it time to log
	if (navSensors[p->portnum].tim - lastLogTime[p->portnum] 
       >= (double) p->interval)
		{
		k = fwrite( (void *) &navSensors[p->portnum], sizeof(t_nav), 1, BinNavLogFP);
		if (k == -1)
			{
			sprintf(buf, "%s log failed %d\n", p->sensor_name, k);
			printf("%s", buf);
			return BAD_FILE_WRITE;
			}
		fflush(BinNavLogFP);
		lastLogTime[p->portnum] = navSensors[p->portnum].tim;
		}

	// check the disk space
	bytesFree = diskFree(drive = getDriveNumber(config->binFilename));

	// check the space remaining
	if ((bytesFree < DISK_FREE_WARNING) &&
      (now - lastWarningTime > DISK_WARNING_INTERVAL))
		{
		lowDiskWarning(bytesFree, drive);
		lastWarningTime = now;
		}
		
	}
return SUCCESS;
}



/*----------------------------------------------------------------------------*/
// Open the Nav file 
/*----------------------------------------------------------------------------*/
int initFileIO(void)
{
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	return BAD_SETUP;
	}

// build full path to filenames
buildFileNames(config);

// Binary data loging
openBinNavLog();

// Ascii data logging
openAscNavLog();

if ((config->BinLoggingEnabled)       ||
    (config->AsciiLogEnabled))
	{
	return SUCCESS;
	}

return LOGGING_NOT_ENABLED;
}

/*----------------------------------------------------------------------------*/
// Close the Binary Nav file 
/*----------------------------------------------------------------------------*/
int closeBinNavLog(void)
{
if (BinNavLogFP != (FILE *) NULL)
	{
	fclose(BinNavLogFP);
	BinNavLogFP = (FILE *) NULL;
	return SUCCESS;
	}
else
	{
	return FILE_NOT_OPEN;
	}
}




/*----------------------------------------------------------------------------*/
// Close the Ascii Nav file 
/*----------------------------------------------------------------------------*/
int closeAscNavLog(void)
{
if (AsciiLogFP != (FILE *) NULL)
	{
	fclose(AsciiLogFP);
	AsciiLogFP = (FILE *) NULL;
	return SUCCESS;
	}
else
	{
	return FILE_NOT_OPEN;
	}
}




/*----------------------------------------------------------------------------*/
// Close the Nav file 
/*----------------------------------------------------------------------------*/
void closeLogFiles(void)
{
closeBinNavLog();
closeAscNavLog();
closeLogbook();
}

/*----------------------------------------------------------------------------*/
// Return a pointer to static Log file pointer
/*----------------------------------------------------------------------------*/
FILE * queryNavLogFile(void)
{
return (FILE * ) BinNavLogFP;
}


/*----------------------------------------------------------------------------*/
// Get Binary File logging name and enable
/*----------------------------------------------------------------------------*/
int doFileLoggingFm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "logdata.frf";
t_setup * config = QuerySetup();
ftag * asciiIntervalTag;
ftag * diskTag;
int binState;
int ascState;
unsigned int logDisk;

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, NULL_SETUP_MSG, frf);
  return FALSE;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

fm->ob_lst[6].flags |= ~IS_EDITABLE;
diskTag = (ftag *) fm->ob_lst[3].tag;
asciiIntervalTag = (ftag *) fm->ob_lst[7].tag;

if (config->logDisk == 0)         // If using the current drive get it
	{
	_dos_getdrive(&logDisk);
	}
else
	{
	logDisk = (unsigned int) config->logDisk;
	}

sprintf(diskTag->fld, "%c:", ('A' + logDisk) - 1);

// set up binary enabled button
if (config->BinLoggingEnabled)
	{
	fm->ob_lst [ 4 ].flags |=  IS_CHECKED;
	}
else
	{
	fm->ob_lst [ 4 ].flags &=  ~IS_CHECKED;
	}

// set up ascii enabled button
if (config->AsciiLogEnabled)
	{
	fm->ob_lst [ 5 ].flags |=  IS_CHECKED;
	}
else
	{
	fm->ob_lst [ 5 ].flags &=  ~IS_CHECKED;
	}

// set up ascii log interval
sprintf(asciiIntervalTag->fld, "%d", config->AsciiLogInterval);

OpenForm ( fm, ob->area.Xmin, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// record the new states of logging
			logDisk  = (unsigned int) (diskTag->fld[0] - 'A') + 1;
			binState = (fm->ob_lst[4].flags & IS_CHECKED) ? 1 : 0;
			ascState = (fm->ob_lst[5].flags & IS_CHECKED) ? 1 : 0;
			config->AsciiLogInterval = atoi(asciiIntervalTag->fld);

			// If log disk has changed, close files and see what gets reopend
			if (logDisk != (unsigned int) config->logDisk)
				{
				config->logDisk           = (int) logDisk;
				closeLogFiles();
				buildFileNames(config);
				}

			config->BinLoggingEnabled = binState;
			config->AsciiLogEnabled   = ascState;

			if (binState)
				{
				// file needs to be open
				openBinNavLog();
				}
			else
				{
				// file has been closed
				closeBinNavLog();
				}

			if (ascState)
				{
				openAscNavLog();
				}
			else
				{
				closeAscNavLog();
				}
			break;

		case 1:                              // CANCEL:
			break;

		case 2:                              // Help
			DisplayAsciiFile("logdata.hlp", 1, 32, 10, 50);
			break;

		case 3:                    // get the disk drive letter
			// if logDisk is the current drive find out what that is and use it
			logDisk = get_disk(diskTag->fld);
			sprintf(diskTag->fld, "%c:", 'A' + logDisk);
			DrawTaggedOb(&fm->ob_lst[3]);
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}






/*----------------------------------------------------------------------------*/
// dump the trackpoint data to a buffer
/*----------------------------------------------------------------------------*/
char * dumpTrackpointVax(t_nav * p, char * buf)
{
int j = 0;

j += sprintf(buf+j, "%d ", p->msmts.usbl.transponder_id);
j += sprintf(buf+j, "%.1lf %.1lf %.1lf", p->msmts.usbl.x,
                                                   p->msmts.usbl.y,
                                                   p->msmts.usbl.z);

j += sprintf(buf+j, " %.1lf", p->msmts.usbl.horiz_angle);
j += sprintf(buf+j, " %.1lf", p->msmts.usbl.slant_range);
j += sprintf(buf+j, " %.1lf", p->msmts.usbl.telemetry);
j += sprintf(buf+j, " %d", p->msmts.usbl.err_codes);

return buf;
}


/*----------------------------------------------------------------------------*/
// open the binary Nav log file
/*----------------------------------------------------------------------------*/
int openBinNavLog(void)
{
int handle;
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, "Yo", NULL_SETUP_MSG);
  return BAD_SETUP;
	}

if (BinNavLogFP != (FILE *) NULL)
	return FILE_ALREADY_OPEN;

if (config->BinLoggingEnabled)
	{
	handle = sopen( config->binFilename, 
                  O_BINARY | O_APPEND | O_RDWR | O_CREAT,
                  SH_DENYNO, S_IREAD | S_IWRITE );

	if (checkCEH())
		return BAD_FILE_OPEN;

	if (handle == -1)
		{
 		NotifyUser(-1,-1,config->binFilename, "Can't open log file" );
		config->BinLoggingEnabled = FALSE;
 		return BAD_FILE_OPEN;
		}

	if ((BinNavLogFP = fdopen(handle, "a")) == (FILE *) NULL)
		{
 		NotifyUser(-1,-1,config->binFilename, "Can't open log file" );
		config->BinLoggingEnabled = FALSE;
 		return BAD_FILE_OPEN;
 		}
	}

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// open the ascii Nav log file
/*----------------------------------------------------------------------------*/
int openAscNavLog(void)
{
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, "Yo", NULL_SETUP_MSG);
  return FALSE;
	}

if (AsciiLogFP != (FILE *) NULL)
	return FILE_ALREADY_OPEN;

if (config->AsciiLogEnabled)
	{
	if ((AsciiLogFP = fopen(config->ascFilename, "a")) == NULL)
  	{
		if (checkCEH())
			return BAD_FILE_OPEN;

  	NotifyUser(-1,-1,config->ascFilename, "Can't open log file" );
		config->AsciiLogEnabled = FALSE;
  	return BAD_FILE_OPEN;
  	}
	}
return SUCCESS;
}






/*----------------------------------------------------------------------------*/
// record the Ascii Nav Data
/*----------------------------------------------------------------------------*/
int recordAsciiNav(void)
{
static time_t lastLogTime = 0L;
time_t now = time(NULL);
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	return BAD_SETUP;
	}

if (config->AsciiLogEnabled)
	{
	// is the log file open
	if (AsciiLogFP == (FILE *) NULL)
		{
		return BAD_FILE_POINTER;
		}

	// is it time to log
	if ((now - lastLogTime) >= (time_t) config->AsciiLogInterval)
		{
		fprintf(AsciiLogFP, "%s %9.5lf %10.5lf %6.1lf%c %5.1lf %5.1lf %5.1lf %s\n", 
            record_time(now), 
            curNav->lat * RAD_TO_DEG, 
            curNav->lon * RAD_TO_DEG, 
            curNav->alt, 
						(depthAlarmFlag) ? '*' : ' ',
            curNav->spd * MPS_TO_KTS, 
            curNav->hdg * RAD_TO_DEG, 
            curNav->crs * RAD_TO_DEG, 
            sensorAbbrevs[curNav->id]);
		fflush(AsciiLogFP);
		lastLogTime = now;
		}
	}

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// Issue a warning message about low disk space
/*----------------------------------------------------------------------------*/
void lowDiskWarning(unsigned long bytesFree, unsigned int drive)
{
char buf[128];
sprintf(buf,
        "Warning: Only %lu bytes left on Disk %c\n", 
        bytesFree, 
        'A' + drive - 1);
Beep();
logMsg(buf);
printMsg(buf);
}


/*----------------------------------------------------------------------------*/
// Build the VAX record
/*----------------------------------------------------------------------------*/
char * buildVaxRecord(t_nav * pNav, char vaxBuf[])
{
int j = 0;
t_com * pCom;
char errBuf[256];

// Check the nav pointer
if (pNav == (t_nav *) NULL)
	{
	sprintf(errBuf, "%s:%d: %s\n", __FILE__, __LINE__, NULL_NAV_MSG);
	scrollMsg(errBuf);
	logMsg(errBuf);
	return (char *) NULL;
	}

// Use the sensor index to get the associated port
pCom = queryComByNav(pNav->id);                  // Sensor Index
if (pCom == (t_com *) NULL)
	{
	scrollMsg(NULL_COM_MSG);
	return (char *) NULL;
	}

switch(pNav->id)
	{
	case TRACKPOINT_II:
		if (pNav->msmts.usbl.transponder_id == 1)
			{
			j+= sprintf(vaxBuf+j, " %s %s %s %s ",
              vaxTime(pNav->tim), 
              vaxLatStr(pNav->lat), 
              vaxLonStr(pNav->lon),
              sensorAbbrevs[pNav->id]);

			strcat(vaxBuf, pCom->buf);
			padBuffer(replaceNewline(vaxBuf), VAX_REC_LENGTH);
			}
		break;

	case DDMU540:
		j = 0;
		j+= sprintf(vaxBuf+j, " %s %s %s ",
              vaxTime(pNav->tim), 
              vaxLatStr(pNav->lat), 
              vaxLonStr(pNav->lon));

		strcat(vaxBuf, pCom->buf);
		replaceNewline(vaxBuf);
		break;

	default:
		return (char *) NULL;
		break;
	}

return (char *) vaxBuf;
}




/*----------------------------------------------------------------------------*/
// build full path to filenames
/*----------------------------------------------------------------------------*/
void buildFileNames(t_setup * config)
{
char drive[_MAX_DRIVE];
char dummy[_MAX_DRIVE];
char dir[_MAX_PATH];
char fname[15];
char ext[15];
unsigned int logDisk;

if (config->logDisk == 0)         // If using the current drive get it
	{
	_dos_getdrive(&logDisk);
	}
else
	{
	logDisk = (unsigned int) config->logDisk;
	}

config->logDisk = (int) logDisk;

// Put all files on One disk
sprintf(drive,"%c:", config->logDisk + 'A' - 1);

// Do the Binary Log File
_splitpath(config->binFilename, dummy, dir, fname, ext);
_makepath(config->binFilename, drive, dir, fname, ext);

// Do the Ascii Log File
_splitpath(config->ascFilename, dummy, dir, fname, ext);
_makepath(config->ascFilename, drive, dir, fname, ext);

// Do the LogBook File
_splitpath(config->logFilename, dummy, dir, fname, ext);
_makepath(config->logFilename, drive, dir, fname, ext);

return;
}


/*----------------------------------------------------------------------------*/
// Change the diskette
/*----------------------------------------------------------------------------*/
int doChangeDiskette ( ob_desc * ob )
{
file_list fList;
t_setup * config = QuerySetup();
char buf[512];
char filename[_MAX_PATH];
int j = 0;
int i = 0;
char drive[_MAX_DRIVE];
char dir[_MAX_PATH];
char fname[15];
char ext[15];
struct _stat filestat;

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, "Yo", NULL_SETUP_MSG);
  return FALSE;
	}

// is this a floppy diskette ??
if (config->logDisk != 1 && config->logDisk != 2)
	{
	NotifyUser(-1, -1, "Yo", "Not Logging to Removable Media");
	}
else
	{
	_splitpath(config->binFilename, drive, dir, fname, ext);
	strcpy(buf, drive);
	strcat(buf, "*.*");
	if ((buildFlist(&fList, buf)) == SUCCESS)
		{
		// Display some file statistics. 
		j += sprintf(buf+j, "\nDisk Contains:\n\n");
		for (i=0; i<fList.nfiles && i < 12; i++)
			{
			j += sprintf(buf+j, "    %s", fList.fnames[i]);
			if( !_stat( fullName(filename, fList.fnames[i]), &filestat ) )
				{
				j+= sprintf(buf+j,  " %ld bytes\n", filestat.st_size );
				}
			else
				{
				j+= sprintf(buf+j,  " ???? bytes\n");
				}
			}
		strcat(buf, "   \n       \n\n");
		DestroyFileList( &fList );
		}

	closeLogFiles();
	NotifyUser( -1, -1, "Remove and Replace Diskette\nPress Okay When Ready", buf );
	logMsg(buf);
	printMsg(buf);
	initFileIO();
	updateDiskFree(TRUE);
	}

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// build a file_list structure
/*----------------------------------------------------------------------------*/
int buildFlist(file_list * fList, char * fileSpec)
{
t_setup * config = QuerySetup();
int j = 0;
int i = 0;
int retv;

Notify( -1, -1, "Scanning directory..." );
retv = CreateFileList( fList, fileSpec, 0 );
NotifyOff();

/* check for possible error returns */
switch( retv )
	{
	case  1 : /* no files found */
		return NO_MATCHING_FILES;
		break;

	case -1 : /* out of memory */
		DestroyFileList( fList );
		NotifyUser(-1, -1, fileSpec, "Out of Memory" );
		return NO_MEMORY;
		break;
	}
return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// return the full path name of the log file
/*----------------------------------------------------------------------------*/
char * fullName(char * buf, char * filename)
{
t_setup * config = QuerySetup();
char drive[_MAX_DRIVE];
char dir[_MAX_PATH];
char fname[15];
char ext[15];
char new_drive[_MAX_DRIVE];
char new_dir[_MAX_PATH];
char new_fname[15];
char new_ext[15];

// get the logfile
_splitpath(config->binFilename, drive, dir, fname, ext);

// get the fname
_splitpath(filename, new_drive, new_dir, new_fname, new_ext);

// substitute the fname
_makepath(buf, drive, dir, new_fname, new_ext);

return buf;
}
