/*----------------------------------------------------------------------------*/
// File: graph.c   USGS Navigation
// Author: John Gann
// Start Date: 13 Feb 1992  Graphics routines for the general navigation
/*----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <limits.h>

#include <GRconst.h>
#include <GRports.h>
#include <GRextrn.h>
#include <menuet.h>

#include "navdefs.h"
#include "navtypes.h"
#include "nav.h"
#include "nav_util.h"
#include "graph.h"
#include "serial.h"
#include "sensor.h"
#include "utmlib.h"
#include "lines.h"
#include "sites.h"
#include "mnutils.h"
#include "stations.h"
#include "video.h"
#include "coast.h"
#include "shapes.h"
#include "plot.h"
#include "offsets.h"

// global data
t_graf graf = {2500L, 0.0, 0.0, 0.0, 0.0, 0, 0};   // global graphics structure
fm_desc * navDispFm = (fm_desc * ) NULL;
fm_desc * rangeFm   = (fm_desc * ) NULL;
int graphWn = -1;
int graphWnWidth;
int graphWnHeight;
rect graphWnRect;
t_shape Vessel;

// Private data
static int graphX;
static int graphY;
static int scrollWn = -1;
static int scrollX;
static int scrollY;
static rect clipRect;
static t_shape Fish;
static int navDispX;
static int navDispY;
static char NewScale = FALSE;
static double shipLat = 0.0;
static double shipLon = 0.0;
static double shipHdg = 0.0;
static double fishLat = 0.0;
static double fishLon = 0.0;
static double fishHdg = 0.0;
static char * CLEAR_MSG = "                                                   ";
static char * CANT_OPEN_MSG = "Can't Open";
static char * NULL_SETUP_MSG  = "Null Configuration";
static char * NULL_COM_MSG    = "Null Com Port";
static char * NULL_TRACK_MSG    = "Null Track Structure";
static char ShipIsDrawn = FALSE;
static char FishIsDrawn = FALSE;

// External data
extern t_com com_struct[];
extern t_nav * curNav;      // pointer to the current navigation 
extern char DisplayingSomething;
extern char RefreshRequired;
extern int postPlotActive;

/*----------------------------------------------------------------------------*/
// initialize the global t_graf structure
/*----------------------------------------------------------------------------*/
void initGraphics(void)
{
t_setup * config = QuerySetup();

read_graf(GRAF_FILE_NAME);
computeGraphRes();
read_shape(&Vessel, config->ShipName);
read_shape(&Fish,   config->FishName);
load_linefile();
load_sitefile();
switch(QuerySetup()->DisplayType)
	{
	case LINE_DISPLAY:
		break;

	case PLAN_DISPLAY:
		setCenter(curNav->lat, curNav->lon);
		break;

	case POLAR_DISPLAY:
		initPolarDisplay();
		break;
	}
}



/*----------------------------------------------------------------------------*/
// Set the Center of the screen to the passed lat lon
/*----------------------------------------------------------------------------*/
int setCenter(double lat, double lon)
{
double scale_factor;
static int computeNewZone = TRUE;
int grid_num;
char grid_ltr;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid for arclength comps

getUTMParam(&grid_num, &grid_ltr);

ll_to_utm(lat, lon,                                    // current pos
          &graf.center_utm.y, &graf.center_utm.x,      // returns UTM xy 
          &scale_factor,                               // returns scale factor
          &grid_num,                                   // returns zone number
          &grid_ltr,                                   // returns zone letter
          USE_PASSED_ZONE,                             // compute new zone
          ellipse->a,                                  // semi major axis
          ellipse->e2);                                // semi minor of WGS84

computeNewZone = FALSE;                                // Only compute zone once
return RefreshRequired = TRUE;
}

/*----------------------------------------------------------------------------*/
// Compute the resolution of pixels to metersin x and y directions
/*----------------------------------------------------------------------------*/
void computeGraphRes(void)
{
graf.xres = (double) graphWnWidth / (double) graf.scale;
graf.yres = (double) graphWnHeight / (double) graf.scale;
graf.centerX = clipRect.Xmin + graphWnWidth / 2;
graf.centerY = clipRect.Ymin + graphWnHeight / 2;
}

/*----------------------------------------------------------------------------*/
// change the current scale by the passed percent
/*----------------------------------------------------------------------------*/
void incrementScale(double change)
{
graf.scale += (long) ((double) graf.scale * change);

if (graf.scale < 50L)
	{
	Beep();
	graf.scale = 50L;
	}

// setCenter(curNav->lat, curNav->lon);
updateScale();
}


/*----------------------------------------------------------------------------*/
// Update the current scale after user has selected through interface
/*----------------------------------------------------------------------------*/
void updateScale(void)
{
computeGraphRes();
RefreshRequired = TRUE;
}

/*----------------------------------------------------------------------------*/
// Convert latitude longitude to screen coordinates
/*----------------------------------------------------------------------------*/
int latLonToScreen(double lat, double lon, int * x, int * y)
{
double scale_factor;
double utmX, utmY;
int retval = TRUE;     // Assume that the coords are within the window
double dx, dy;
double half_scale =  .9 * ((double) graf.scale / 2.0);
long lx, ly;
int grid_num;
char grid_ltr;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid for arclength comps

getUTMParam(&grid_num, &grid_ltr);

// get the UTM coordinate of the passed lat and lon
ll_to_utm(lat, lon,                                    // current pos
          &utmY, &utmX,                                // returns UTM xy 
          &scale_factor,                               // returns scale factor
          &grid_num,                                   // returns zone number
          &grid_ltr,                                   // returns zone letter
          USE_PASSED_ZONE,                             // compute new zone
          ellipse->a,                                  // semi major axis
          ellipse->e2);                                // semi minor of WGS84

// compute the offset from the center of the screen
dx = graf.center_utm.x - utmX;
dy = graf.center_utm.y - utmY;

// check the deltas to see if we are within the window
if ((fabs(dx) > half_scale) || (fabs(dy) > half_scale))
	retval = FALSE;                                     // set up the return val

// convert the offset to screen coordinates in pixels

lx = (long) (dx * -graf.xres) + (long) graf.centerX;
ly = (long) (dy *  graf.yres) + (long) graf.centerY;

// make sure they are still integers before assigning them
if (lx < (long) INT_MIN)
	lx = (long) INT_MIN;

if (lx > (long) INT_MAX)
	lx = (long) INT_MAX;

if (ly < (long) INT_MIN)
	ly = (long) INT_MIN;

if (ly > (long) INT_MAX)
	ly = (long) INT_MAX;

*x = (int) lx;
*y = (int) ly;

return retval;
}


/*----------------------------------------------------------------------------*/
// Convert UTM X and Y to screen coords
/*----------------------------------------------------------------------------*/
int utmToScreen(double utmX, double utmY, int * x, int * y)
{
int retval = TRUE;     // Assume that the coords are within the window
double dx, dy;
double half_scale =  .9 * ((double) graf.scale / 2.0);
long lx, ly;

// compute the offset from the center of the screen
dx = graf.center_utm.x - utmX;
dy = graf.center_utm.y - utmY;

// check the deltas to see if we are within the window
if ((fabs(dx) > half_scale) || (fabs(dy) > half_scale))
	retval = FALSE;                                     // set up the return val

// convert the offset to screen coordinates in pixels
lx = (long) (dx * -graf.xres) + (long) graf.centerX;
ly = (long) (dy *  graf.yres) + (long) graf.centerY;

// make sure they are still integers before assigning them
if (lx < (long) INT_MIN)
	lx = (long) INT_MIN;

if (lx > (long) INT_MAX)
	lx = (long) INT_MAX;

if (ly < (long) INT_MIN)
	ly = (long) INT_MIN;

if (ly > (long) INT_MAX)
	ly = (long) INT_MAX;

*x = (int) lx;
*y = (int) ly;

return retval;
}



/*----------------------------------------------------------------------------*/
// Convert screen x and y to lat and lon
// x and y are global screen coordinates i.e. from 0-640 and 0-480
/*----------------------------------------------------------------------------*/
int screenToLatLon(int x, int y, double * lat, double * lon)
{
double half_scale =  (double) graf.scale / 2.0;
int dxPix;             // offset from center in pixels
int dyPix;             // offset from center in pixels
double N, E;           // Northing and Easting in meters
int retval = TRUE;     // Assume that the coords are within the window
int grid_num;
char grid_ltr;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid for arclength comps

getUTMParam(&grid_num, &grid_ltr);

// compute the offset in pixels from the center of the window
dxPix = graf.centerX - x;
dyPix = graf.centerY - y;

// See if the passed point was even on the screen
retval = MouseInDrawingArea(x, y);

// convert the pixel offset to meters
E = (double) dxPix / -graf.xres + graf.center_utm.x;
N = (double) dyPix /  graf.yres + graf.center_utm.y;

// Compute the lat lon coordinate of the utm location
utm_to_ll(N, E,                                  // passed y and x
          lat, lon,                              // returned position
          grid_num,                              // grid zone number
          grid_ltr,                              // grid zone letter
          ellipse->a,                            // semi maj axis
          ellipse->e2);                          // eccentricity squared

return retval;
}



/*----------------------------------------------------------------------------*/
// Build Nav display form and window
/*----------------------------------------------------------------------------*/
int buildNavDisplay(void)
{
int full_width, full_height;
int fm_width, fm_height;                             // form dimensions
rect work;
rect titleRect;
char buf[256];
int retv;
int commentHeight = 0;
mn_win_rec * wn_rec;
attr_blk * comm_blk;
attr_blk * win_blk;

if ( ( navDispFm = loadSidePanelFRF() ) == (fm_desc *) NULL ) 
	{
	return FALSE;
	}

DupRect ( QryWnRect ( 0, WN_WORK_RECT ), &work );
full_height = work.Ymax - work.Ymin;
full_width = work.Xmax - work.Xmin;

fm_width  = navDispFm->area.Xmax - navDispFm->area.Xmin;
fm_height = navDispFm->area.Ymax - navDispFm->area.Ymin;

// compute area of graphic window
graphWnWidth = full_width - fm_width;
graphWnHeight = full_height;
navDispX = graphWnWidth;
navDispY = work.Ymin;

if (!(retv = OpenForm( navDispFm, navDispX, navDispY)))
	{
	sprintf(buf, "%s %s at line %d retv = %d navDispX = %d navDispY = %d\n", CANT_OPEN_MSG, __FILE__, __LINE__, retv, navDispX, navDispY);
	Puke(buf);
	}

// give the graphics area a different border 
win_blk = GetAttrBlk( attrUsr4 );
win_blk->bdr_style = BDR_STACKED;
win_blk->thk = 1;
SetWnBdrAttr( attrUsr4 );
SetBdrColors( 7, 1 );

// Create Graphics area
if ((graphWn = CreateWindow(NULL, NULL, CLEAR_MSG, 
                               graphWnWidth, full_height)) == -1)
	{
	printf("Unable to Create the Graphic Window\n");
	return FALSE;
	}

DupRect ( QryWnRect ( 0, WN_WORK_RECT ), &work );
DupRect ( QryWnRect ( 0, WN_TITL_RECT ), &titleRect );


graphY = work.Ymin;
graphX = work.Xmin;
if (!OpenWindow(graphWn, graphX,  graphY))
	{
	printf("Unable to Open the Graphic Window\n");
	return FALSE;
	}
else
	{
	wn_rec = QryWnRec(graphWn);
	comm_blk = GetAttrBlk(wn_rec->attr_c);
	commentHeight = (wn_rec->wn_comm.Ymax - wn_rec->wn_comm.Ymin) + 
                   4 * comm_blk->thk;

	SetRect(&clipRect,    comm_blk->thk + 2, 
                        comm_blk->thk + 1, 
                        graphX + graphWnWidth - (3 * (comm_blk->thk + 1)), 
                        graphY + full_height - 
                                 2 * commentHeight - 
                                 comm_blk->thk + 1);

	SetRect(&graphWnRect, graphX,  
                        graphY + comm_blk->thk, 
                        graphX + (graphWnWidth - comm_blk->thk * 2), 
                        graphY + (full_height - commentHeight));

	graphWnHeight = clipRect.Ymax - clipRect.Ymin;
	graphWnWidth  = clipRect.Xmax - clipRect.Xmin;
	SetActiveWn(0, 0);
	// NotifyUser(-1,-1, "clipRect", dumpRect(&clipRect));
	// NotifyUser(-1,-1, "graphWnRect", dumpRect(&graphWnRect));
	}

// Create Scroll area
if ((scrollWn = CreateWindow(NULL, NULL, NULL, 
      fm_width, (full_height - fm_height) - 8)) == -1)
	{
	printf("Unable to Create the Graphic Window\n");
	return FALSE;
	}

DupRect ( QryWnRect ( 0, WN_WORK_RECT ), &work );
scrollY = navDispFm->area.Ymax;
scrollX = navDispFm->area.Xmin;
if (!OpenWindow(scrollWn, scrollX,  scrollY))
	{
	printf("Unable to Open the Scroll Window\n");
	return FALSE;
	}
else
	{
	SetActiveWn(0, 0);
	}

// Set up the global graphics structure
initGraphics();

return SUCCESS;
}



/*----------------------------------------------------------------------------*/
/* Save the graphics setup
/*----------------------------------------------------------------------------*/
int write_graf(char * filename)
{
FILE *fp;
int k;

if ((fp = fopen(filename, "wb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  fclose(fp);
  return FALSE;
  }

k = fwrite( (char *) &graf, sizeof(t_graf), 1, fp);
if (k == -1)
	{
	fclose(fp);
	printf("\nfwrite failed");
	return FALSE;
	}

fclose(fp);
return TRUE;
}



/*----------------------------------------------------------------------------*/
// read the graphics set
/*----------------------------------------------------------------------------*/
int read_graf(char * filename)
{
FILE *fp;
int k;

if ((fp = fopen(filename, "rb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  fclose(fp);
  return FALSE;
  }

k = fread( (char *) &graf, sizeof(t_graf), 1, fp);
if (k == -1)
	{
	fclose(fp);
	printf("\nfread failed");
	return FALSE;
	}

// dumpGraphics();

fclose(fp);
return TRUE;
}






/*----------------------------------------------------------------------------*/
// display the current graf struct
/*----------------------------------------------------------------------------*/
void dumpGraphics(void)
{
int j = 0;
char buf[512];

j += sprintf(buf+j, "scale = %ld\n", graf.scale);
j += sprintf(buf+j, "center_utm.x = %.2lf\n", graf.center_utm.x);
j += sprintf(buf+j, "center_utm.y = %.2lf\n", graf.center_utm.y);
j += sprintf(buf+j, "xres = %.5lf\n", graf.xres);
j += sprintf(buf+j, "yres = %.5lf\n", graf.yres);
j += sprintf(buf+j, "centerX = %d\n", graf.centerX);
j += sprintf(buf+j, "centerY = %d\n", graf.centerY);
j += sprintf(buf+j, "graphWnHeight = %d\n", graphWnHeight);
j += sprintf(buf+j, "graphWnWidth = %d\n", graphWnWidth);

NotifyUser(-1, -1, GRAF_FILE_NAME, buf);
}




/*----------------------------------------------------------------------------*/
// display the current graf struct
/*----------------------------------------------------------------------------*/
void dumpObs(t_observations * obs)
{
int j = 0;
int i;
char buf[512];
int lastWn = QryActiveWn();

for (i=0; i<obs->numObs; i++)
	{
	j += sprintf(buf+j, "%s %d %.1lf %.1lf %.1lf\n", 
             obs->stations[i]->name, 
             obs->stations[i]->id, 
             obs->ranges[i], 
             obs->residuals[i], 
             RAD_TO_DEG * obs->bearings[i]);
	}

SetActiveWn(graphWn, 0);
j += sprintf(buf+j, "Iterations = %d\n", obs->iterations);
j += sprintf(buf+j, "numObs = %d\n\n", obs->numObs);
SetActiveWn(lastWn, 0);

printWn(buf);
}



/*----------------------------------------------------------------------------*/
// Draw the Ship
/*----------------------------------------------------------------------------*/
void drawShip(double lat, double lon, int color)
{
static int last_x = 0, last_y = 0;
static double last_hdg = 0.0;
int x, y;
double hdg;
t_setup * config = QuerySetup();
t_ellips * ellipse;
double distance, fazm, bazm;
t_pos pos;

pos.lat = lat;
pos.lon = lon;
if (validPos(&pos) != SUCCESS)
	return;

SetClipArea();

pushGrEnv();
PenMode(zNXORz);
// PenSize( 1, 1 );
// PenPattern( 1 );
// PenDash( 0 );
// HideCursor();
// PenColor(color);

if (!latLonToScreen(lat, lon, &x, &y))         // it ain't on the screen
	{
	if (config->misc.shipCentering)
		{
		switch(QuerySetup()->DisplayType)
			{
			case PLAN_DISPLAY:
				setCenter(lat, lon);
				break;

			case POLAR_DISPLAY:
				// Ship won't fit on screen at this scale so find new scale 
        ellipse = query_ellipsoid();
				arclen(config->curSite.pos.lat, config->curSite.pos.lon,
               curNav->lat, curNav->lon,
               &distance,
               &fazm,
               &bazm,
               ellipse->a,
               ellipse->b);
				graf.scale = (long) (distance + distance * .10) * 2L;
				updateScale();
				setCenter(config->curSite.pos.lat, config->curSite.pos.lon);
				break;
			}
		latLonToScreen(lat, lon, &x, &y);
		}
	}

// if the ship has not moved and the gyro has not changed returned
hdg = getHeading();
eraseShip(color);

drawRotatableObject(&Vessel, hdg, x, y, config->misc.shipScaling);

shipLat = lat;
shipLon = lon;
shipHdg = hdg;
	
last_x    = x;
last_y    = y;
last_hdg  = hdg;
ShipIsDrawn = TRUE;
popGrEnv();

SetActiveWn(0, 0);
PenSize( 0, 0 );
ShowCursor();
}



/*----------------------------------------------------------------------------*/
// Erase the Ship
/*----------------------------------------------------------------------------*/
void eraseShip(int color)
{
int x, y;
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();

if (!ShipIsDrawn)
	return;

if (shipLat != 0.0 && shipLon != 0.0)
	{
	SetActiveWn(graphWn, 0);
	SetClipArea();
	pushGrEnv();

	PenMode(zNXORz);
	// PenSize( 1, 1 );
	// PenPattern( 1 );
	// PenDash( 0 );
	// PenColor( color );
	// HideCursor();
	if (!latLonToScreen(shipLat, shipLon, &x, &y))
		{
		ShowCursor();
		popGrEnv();
		return;                       // don't try to erase if it's not on screen
		}
	drawRotatableObject(&Vessel, shipHdg, x, y, config->misc.shipScaling);
	popGrEnv();
	ShowCursor();
	SetActiveWn(lastWn, 0);
	}
}



/*----------------------------------------------------------------------------*/
// Draw an indexed shape
/*----------------------------------------------------------------------------*/
void draw_symbol ( int symbol, int x, int y )
{
rect r;

switch ( symbol ) 
	{
	case DOT : /* Dot */
		SetPixel ( x, y );
		break;

	case CROSS : /* Cross */
		MoveTo ( x-3, y );
		LineTo ( x+3, y );
		MoveTo ( x, y-3 );
		LineTo ( x, y+3 );
		break;

	case THE_X : /* 'X' */
		MoveTo ( x-3, y-3 );
		LineTo ( x+3, y+3 );
		MoveTo ( x-3, y+3 );
		LineTo ( x+3, y-3 );
		break;

	case SQUARE_DOT : /* Square Dot */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameRect ( &r );
		SetPixel ( x, y );
		break;

	case CIRCLE_DOT : /* Circle Dot */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameOval ( &r );
		SetPixel ( x, y );
		break;

	case TRIANGLE_DOT : /* Triangle Dot */
		MoveTo ( x-6, y+3 );
		LineTo ( x,   y-7 );
		LineTo ( x+6, y+3 );
		LineTo ( x-6, y+3 );
		SetPixel ( x, y );
		break;

	case LIGHT_POLE : /* Light Pole */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameOval ( &r );
		MoveTo ( x-4, y-4 );
		LineTo ( x-6, y-6 );
		MoveTo ( x+4, y-4 );
		LineTo ( x+6, y-6 );
		MoveTo ( x+4, y+4 );
		LineTo ( x+6, y+6 );
		MoveTo ( x-4, y+4 );
		LineTo ( x-6, y+6 );
		break;
         
	case FIRE_HYDRANT : /* Fire Hydrant */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameOval ( &r );
		MoveTo ( x-5, y );
		LineTo ( x-8, y );
		MoveTo ( x, y-5 );
		LineTo ( x, y-8 );
		MoveTo ( x+5, y );
		LineTo ( x+8, y );
		break;

	case CROSSED_SQUARE : /* Crossed Square */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameOval ( &r );
		MoveTo ( x, y-5 );
		LineTo ( x, y+5 );
		MoveTo ( x-5, y );
		LineTo ( x+5, y );
		break;
         
	case CROSSED_CIRCLE : /* Crossed Circle */
		SetRect ( &r, x-5, y-5, x+5, y+5 );
		FrameRect ( &r );
		MoveTo ( x, y-10 );
		LineTo ( x, y+10 );
		MoveTo ( x-10, y );
		LineTo ( x+10, y );
		break;

	case THING_A_MA_BOB : 
		MoveTo ( x-1,  y-12 );
		LineTo ( x-1,  y-3  );
		MoveTo ( x+1,  y-12 );
		LineTo ( x+1,  y-3  );
		MoveTo ( x-1,  y+12 );
		LineTo ( x-1,  y+3  );
		MoveTo ( x+1,  y+12 );
		LineTo ( x+1,  y+3  );
		MoveTo ( x-12, y-1  );
		LineTo ( x-3,  y-1  );
		MoveTo ( x-12, y+1  );
		LineTo ( x-3,  y+1  );
		MoveTo ( x+12, y-1  );
		LineTo ( x+3,  y-1  );
		MoveTo ( x+12, y+1  );
		LineTo ( x+3,  y+1  );
		break;
         
case HORIZ_BRIDGE:
		MoveTo ( x-3,  y-2 );
		LineTo ( x-2,  y-1 );
		LineTo ( x+2,  y-1 );
		LineTo ( x+3,  y-2 );
		MoveTo ( x-3,  y+2 );
		LineTo ( x-2,  y+1 );
		LineTo ( x+2,  y+1 );
		LineTo ( x+3,  y+2 );
		SetPixel ( x, y );
		break;

case VERT_BRIDGE:
		MoveTo ( x-2,  y-3 );
		LineTo ( x-1,  y-2 );
		LineTo ( x-1,  y+2 );
		LineTo ( x-2,  y+3 );
		MoveTo ( x+2,  y-3 );
		LineTo ( x+1,  y-2 );
		LineTo ( x+1,  y+2 );
		LineTo ( x+2,  y+3 );
		SetPixel ( x, y );
		break;

case THE_Y:
		MoveTo ( x-2,  y-2 );
		LineTo ( x,  y );
		LineTo ( x+2,  y-2 );
		MoveTo ( x,  y );
		LineTo ( x,  y+3 );
		break;

case FARNELLA:
		MoveTo ( x,   y-32);
		LineTo ( x+4, y-22);
		LineTo ( x+6, y-8);
		LineTo ( x+6, y+32);
		LineTo ( x+2, y+32);
		LineTo ( x+2, y+30);
		LineTo ( x-2, y+30);
		LineTo ( x-2, y+32);
		LineTo ( x-6, y+32);
		LineTo ( x-6, y-8);
		LineTo ( x-4, y-22);
		LineTo ( x,   y-32);
		break;

	default : /* Dot */
		SetPixel ( x, y );
		break;
}

MoveTo ( x, y );
}



/*----------------------------------------------------------------------------*/
// Draw a rotatable object
/*----------------------------------------------------------------------------*/
void drawRotatableObject ( t_shape * ob, 
                           double hdg, 
                           int x, int y, 
                           int isScaled )
{
int X, Y;
int i;
double xres = (isScaled) ? graf.xres : 1.0;
double yres = (isScaled) ? graf.yres : 1.0;
point pt1, pt2;

// rotate the heading for drawing purposes
hdg = rotateBrg(hdg, -90.0 * DEG_TO_RAD);

for(i=0; i< ob->numEntities; i++)
	{
	switch(ob->entities[i].type)
		{
		case G_POINT:
			X = (int) (xres * ob->entities[i].shape.point.v.magnitude * cos(ob->entities[i].shape.point.v.direction + hdg)) + x;
			Y = (int) (yres * ob->entities[i].shape.point.v.magnitude * sin(ob->entities[i].shape.point.v.direction + hdg)) + y;
			// Y += 32;
			drawPointEntity(&ob->entities[i], X, Y);
			break;

		case G_LINE:
			pt1.X = (int) (xres * ob->entities[i].shape.line.v1.magnitude * cos(ob->entities[i].shape.line.v1.direction + hdg)) + x;
			pt1.Y = (int) (yres * ob->entities[i].shape.line.v1.magnitude * sin(ob->entities[i].shape.line.v1.direction + hdg)) + y;
			// pt1.Y += 32;
			pt2.X = (int) (xres * ob->entities[i].shape.line.v2.magnitude * cos(ob->entities[i].shape.line.v2.direction + hdg)) + x;
			pt2.Y = (int) (yres * ob->entities[i].shape.line.v2.magnitude * sin(ob->entities[i].shape.line.v2.direction + hdg)) + y;
			// pt2.Y += 32;
			drawLineEntity(&ob->entities[i], &pt1, &pt2);
			break;

		case G_OVAL:
			pt1.X = (int) (xres * ob->entities[i].shape.oval.v.magnitude * cos(ob->entities[i].shape.oval.v.direction + hdg)) + x;
			pt1.Y = (int) (yres * ob->entities[i].shape.oval.v.magnitude * sin(ob->entities[i].shape.oval.v.direction + hdg)) + y;
			// pt1.Y += 32;
			pt2.X = (int) (xres * ob->entities[i].shape.oval.xRadius);  
			pt2.Y = (int) (yres * ob->entities[i].shape.oval.yRadius); 
			drawOvalEntity(&ob->entities[i], &pt1, &pt2);
			break;
		}
	}
}

/*----------------------------------------------------------------------------*/
// Draw a point entity using the display features
/*----------------------------------------------------------------------------*/
void drawPointEntity(t_gEntity * pE, int X, int Y)
{
pushGrEnv();

PenColor(pE->shape.point.display.penColor);           // Set up graphics
ClipRect(&clipRect);

MoveTo(X, Y-pE->shape.point.display.penSize);         // Draw it
LineTo(X, Y+pE->shape.point.display.penSize);         // Draw it
MoveTo(X-pE->shape.point.display.penSize, Y);         // Draw it
LineTo(X+pE->shape.point.display.penSize, Y);         // Draw it
   
popGrEnv();                               // Restore Graphics
}


/*----------------------------------------------------------------------------*/
// Draw a line entity using the display features
/*----------------------------------------------------------------------------*/
void drawLineEntity(t_gEntity * pE, point * pt1, point * pt2)
{
pushGrEnv();                            // Preserve Graphics
PenDash (pE->shape.line.display.lineType);    // Set up line style
PenSize (pE->shape.line.display.penSize,      // Set up pen width
         pE->shape.line.display.penSize);     // Set up pen width
PenColor(pE->shape.line.display.penColor);    // Set pen color
ClipRect(&clipRect);

MoveTo(pt1->X, pt1->Y);                 // Draw it
LineTo(pt2->X, pt2->Y);                 // Draw it

popGrEnv();                             // Restore Graphics
}


/*----------------------------------------------------------------------------*/
// Draw a oval entity using the display features
// pt1 contains the center of oval
// pt2 contains the x and y radii respectively
/*----------------------------------------------------------------------------*/
void drawOvalEntity(t_gEntity * pE, point * pt1, point * pt2)
{
rect r;

pushGrEnv();                                       // Preserve Graphics
PenDash (pE->shape.oval.display.lineType);       // Set up line style
PenSize (pE->shape.oval.display.penSize,         // Set up pen width
         pE->shape.oval.display.penSize);        // Set up pen width
PenColor(pE->shape.oval.display.penColor);       // Set pen color
ClipRect(&clipRect);

r.Xmin = pt1->X - pt2->X;                          // build up the inscribed rct
r.Xmax = pt1->X + pt2->X;                          // build up the inscribed rct
r.Ymin = pt1->Y - pt2->Y;                          // build up the inscribed rct
r.Ymax = pt1->Y + pt2->Y;                          // build up the inscribed rct

if (pE->shape.oval.display.fillAble)             // Fill it if necessary
	{
	PenPattern(pE->shape.oval.display.fillPattern);    // Set fill color
	PenColor(pE->shape.oval.display.fillColor);    // Set fill color
	PaintOval(&r);
	}

if (pE->shape.oval.display.frameAble)            // Frame it if necessary
	{
	PenPattern(1);
	PenColor(pE->shape.oval.display.frameColor);   // Set frame color
	FrameOval(&r);
	}

popGrEnv();                             // Restore Graphics
}



/*----------------------------------------------------------------------------*/
// do all the graphics drawing stuff here
/*----------------------------------------------------------------------------*/
int doGraphics(void)
{
t_pos * vrpPos;
t_nav * p = (t_nav *) NULL;

if (!OKtoDraw())
	return -2;

switch(QuerySetup()->DisplayType)
	{
	case PLAN_DISPLAY:
	case LINE_DISPLAY:
	case POLAR_DISPLAY:
		vrpPos = queryVRP();
		drawShip(vrpPos->lat, vrpPos->lon, curNav->id);
		drawTrailLines();
		drawFish();
		break;

	case RANGE_REDUCTION_DISPLAY:
		p = queryNavByType(DDMU540);
		if (p != (t_nav *) NULL)
			{
			// dumpObs(&p->msmts.rng.obs);
			updateRangeForm(p);
			}
		break;
	}

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// Return TRUE if no forms are open && no dropdown menus are active
/*----------------------------------------------------------------------------*/
int OKtoDraw(void)
{
fm_desc * menu = QueryMenu();
fm_desc * thisFm;
int i;

for (i=0; i<MENU_PICKS; i++)
	{
	thisFm = (fm_desc *) menu->ob_lst[i].tag;            
	if (thisFm->flags & IS_OPEN)           // file menu
		return FALSE;
	}

// if (postPlotActive && isScreenPlot())
	// return FALSE;

return TRUE;
}



/*----------------------------------------------------------------------------*/
// draw all the lines if they are enabled for display
/*----------------------------------------------------------------------------*/
void drawLines(void)
{
int max_lines = getLineCount();
int i;
int width;
t_setup * config = QuerySetup();
t_line * thisLine;

if (config == (t_setup *) NULL)
	{
	scrollMsg(NULL_SETUP_MSG);
  return;
	}

for (i=0; i<max_lines; i++)
	{
	// get the next line
	thisLine = getNthLine(i);
	if (thisLine == (t_line *) NULL)
		{
		NotifyUser(-1, -1, "Error", "Unable to match line\n");
		}

	if (thisLine->display)
		{
		// highlight the current line
		if (!memcmp((void *) &config->curLine,
              	(void *) thisLine, sizeof(t_line)))
			width = 3;
		else
			width = 1;
	
		drawLine(thisLine, width);
		}
	}
}

/*----------------------------------------------------------------------------*/
// draw the lines at the specified pixel width
/*----------------------------------------------------------------------------*/
void drawLine(t_line * line, int width)
{
int x1, y1, x2, y2;

if (line == (t_line *) NULL)
	return;

SetActiveWn(graphWn, 0);
SetClipArea();
PenSize( width, width );
PenMode(zNXORz);
PenPattern( 1 );
PenDash( 0 );
HideCursor();
PenColor(GREEN);

latLonToScreen(line->start.lat, line->start.lon, &x1, &y1);
latLonToScreen(line->end.lat, line->end.lon, &x2, &y2);

MoveTo(x1, y1);
LineTo(x2, y2);
textAnnot(line->name, x2, y2, 5, 7);

PenSize( 0, 0 );
SetActiveWn(0, 0);
ShowCursor();
PenColor(BLACK);

}

/*----------------------------------------------------------------------------*/
// draw all the graphical objects 
/*----------------------------------------------------------------------------*/
int refreshGraphics(void)
{
int lastWn = QryActiveWn();

if (!OKtoDraw())
	return -1;

SetActiveWn(0, 0);

switch(QuerySetup()->DisplayType)
	{
	case LINE_DISPLAY:
		killRangeForm();        // if it's alive
		ClearGraphicsArea();
		drawHistory(queryShipHistory());
		break;

	case PLAN_DISPLAY:
		HideCursor();
		killRangeForm();        // if it's alive
		ClearGraphicsArea();
		drawLines();
		drawSites();
		drawStations();
		drawCoastline(queryCoastline());
		drawHistory(queryShipHistory());
		ShowCursor();
		break;

	case POLAR_DISPLAY:
		HideCursor();
		killRangeForm();        // if it's alive
		ClearGraphicsArea();
		initPolarDisplay();
		drawLines();
		drawSites();
		drawStations();
		drawCoastline(queryCoastline());
		drawHistory(queryShipHistory());
		ShowCursor();
		break;

	case RANGE_REDUCTION_DISPLAY:
		break;
	}

// redraw videoWn
refreshVideo();

NewScale = TRUE;
SetActiveWn(lastWn, 0);
return SUCCESS;
}



/*----------------------------------------------------------------------------*/
// scroll message in a window 
/*----------------------------------------------------------------------------*/
void scrollMsg(char * buf)
{
int lastWn = QryActiveWn();

// check for NULL pointer
if (buf == (char *) NULL)
	return;

// Make sure the window is open and safe to display
if (scrollWn == -1)
	return;

SetActiveWn ( scrollWn, 0 );		
printWn ( buf );
SetActiveWn ( lastWn, 0 );
}



/*----------------------------------------------------------------------------*/
// Get the heading based on the current gyro
// Heading in radians is returned if available else crs made good is returned
/*----------------------------------------------------------------------------*/
double getHeading(void)
{
t_com * gyroCom = (t_com *) NULL;       // pointer to Gyro port struct
t_nav * gyroNav = (t_nav *) NULL;     // pointer to Gyro nav struct
// char buf[128];
double hdg = -1.0;

gyroCom  = queryComByNav(QuerySetup()->Gyro);   // get a pointer to Gyro port
if (gyroCom == (t_com *) NULL)
	{
	// must not be a gyro connected so use the crs made good
	hdg = curNav->crs;
	}
else
	{
	// you have the com structure so get the nav structure
	gyroNav = queryNavPtr(gyroCom->portnum);
	if (gyroNav == (t_nav *) NULL)
		{
	// 	sprintf(buf, "Null nav at %d of %s\n", __LINE__, __FILE__);
		// scrollMsg(buf);
		// There is a problem
		hdg = curNav->crs;
		}
	else
		{
		hdg = gyroNav->hdg;
		}
	}
return hdg;
}


/*----------------------------------------------------------------------------*/
// See if Pixel X and Y are within the drawing area
/*----------------------------------------------------------------------------*/
int MouseInDrawingArea(int x, int y)
{
point pt;
pt.X = x;
pt.Y = y;
return (PtOnRect(&pt, &graphWnRect, 1, 1));
}


/*----------------------------------------------------------------------------*/
// Update the cursor coordinates
/*----------------------------------------------------------------------------*/
void updateCursorCoords(int x, int y, int forceUpdate)
{
static int lastX, lastY;
char latBuf[30];
char lonBuf[30];
char altBuf[30];
char buf[80];
double lat = 0.0, lon = 0.0;
int lastWn = QryActiveWn();

// See if its's necessary to update the mouse coordinates
if (forceUpdate || lastX != x || lastY != y)
	{
	SetActiveWn(graphWn, 0);
	if (MouseInDrawingArea(x, y))
		{
		screenToLatLon(x - graphWnRect.Xmin, y - graphWnRect.Ymin, &lat, &lon);
		PositionString(lonBuf, latBuf, altBuf, 
                   lon, lat, 0.0, 0);
		strcpy(buf, latBuf);
		strcat(buf, "   ");
		strcat(buf, lonBuf);
		strcat(buf, " (Scale: ");
		strcat(buf, ltoa(graf.scale, altBuf, 10));
		strcat(buf, " )");
		DisplayComment(buf);
		}	
	else
		{
		DisplayComment(CLEAR_MSG);
		}
	
	lastX = x;
	lastY = y;
	SetActiveWn(lastWn, 0);
	}
}

/*----------------------------------------------------------------------------*/
// Pick from a list of Scales
/*----------------------------------------------------------------------------*/
int doPickScale ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "pickScal.frf";
ltag * tag;
char buf[80];

/* Load the FRF file */
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

tag = (ltag *) fm->ob_lst[2].tag;
sprintf(buf, "%ld", graf.scale);
tag->curitem = get_ltag_match(buf, tag);

OpenForm ( fm, ob->area.Xmin-100, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the program index
			tag = (ltag *) fm->ob_lst[2].tag;
			graf.scale = atol(tag->list[tag->curitem]);
			updateScale();
			break;

		case CANCEL-1:
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL-1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}


/*----------------------------------------------------------------------------*/
// Pick scale using rubber rectangle
/*----------------------------------------------------------------------------*/
int doRubberRect ( ob_desc * ob )
{
int retv;
rect r;
int lastWn = QryActiveWn();
double leftLon, rightLon, topLat, botLat, centerLat, centerLon;
double distance, fazm, bazm;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid for arclength comps

// point to the graphics window
SetActiveWn(graphWn, 0);
DisplayComment(" Draw a RECT, ESC to Abort " );

// draw the rubber rect
retv = GetRubRect( QryWnRect(graphWn, WN_CLIENT_AREA), &r );

// check rctangle limits to prevent floating point crash
if ((r.Xmax == r.Xmin) || (r.Ymax == r.Ymin))
	{
	Beep();
	return RUBBER_SHAPE_LIMIT_ERR;
	}

// compute the new scale
if (retv)
	{
	screenToLatLon(r.Xmin, r.Ymin, &topLat, &leftLon);
	screenToLatLon(r.Xmax, r.Ymax, &botLat, &rightLon);
	centerLat = (botLat + topLat) / 2.0;
	centerLon = (leftLon + rightLon) / 2.0;
	arclen(topLat, leftLon,
         botLat, rightLon,
         &distance,
         &fazm,
         &bazm,
         ellipse->a,
         ellipse->b);
	graf.scale = (long) distance;
	updateScale();
	setCenter(centerLat, centerLon);
	}

// clean up
DisplayComment(CLEAR_MSG);
SetActiveWn(lastWn, 0);

return retv;
}




/*----------------------------------------------------------------------------*/
// set Center using mouse
/*----------------------------------------------------------------------------*/
int doSetCenter ( ob_desc * ob )
{
int retv;
int lastWn = QryActiveWn();
double lon, lat;
time_t now = time(NULL);
time_t abortTime = now + 10L;
point * p;

// point to the graphics window
SetActiveWn(graphWn, 0);
DisplayComment(" Click on desired screen center");

// loop until user clicks or timeout occurs
while(abortTime - now > 0L)
	{
	now = time(NULL);
	if (retv = MsgInQueue())
		{
		// compute the new center
		p = MouseXY(0);
		screenToLatLon(p->X, p->Y, &lat, &lon);
		setCenter(lat, lon);
		// clean up
		DisplayComment(CLEAR_MSG);
		SetActiveWn(lastWn, 0);
		return retv;
		}
	}

// clean up
DisplayComment(CLEAR_MSG);
SetActiveWn(lastWn, 0);

return retv;
}







/*----------------------------------------------------------------------------*/
// draw all the stations if they are enabled for display
/*----------------------------------------------------------------------------*/
void drawStations(void)
{
t_station * thisStation;
int i;

for (i=0; i<MAX_NAV_STATIONS; i++)
	{
	// get the next station
	thisStation = queryStation(i);
	if (thisStation == (t_station *) NULL)
		{
		NotifyUser(-1, -1, "Error", "Unable to match thisStation\n");
		return;
		}

	if (thisStation->enabled && thisStation->display)
		drawStation(thisStation);
	}
}

/*----------------------------------------------------------------------------*/
// draw all the stations if they are enabled for display
/*----------------------------------------------------------------------------*/
void drawStation(t_station * thisStation)
{
int x, y;

if (thisStation == (t_station *) NULL)
	return;

SetActiveWn(graphWn, 0);
SetClipArea();
PenSize( 1, 1 );
PenMode(zNXORz);
PenPattern( 1 );
PenDash( 0 );
HideCursor();
PenColor(RED);

latLonToScreen(thisStation->pos.lat, thisStation->pos.lon, &x, &y);
draw_symbol(TRIANGLE_DOT, x, y);

PenSize( 0, 0 );
SetActiveWn(0, 0);
ShowCursor();
PenColor(BLACK);
}


/*----------------------------------------------------------------------------*/
// draw all the sites if they are enabled for display
/*----------------------------------------------------------------------------*/
void drawSites(void)
{
int maxSites = getSiteCount();
int i;
int width;
t_setup * config = QuerySetup();
t_site * thisSite;

if (config == (t_setup *) NULL)
	{
	scrollMsg(NULL_SETUP_MSG);
  return;
	}

for (i=0; i<maxSites; i++)
	{
	// get the next line
	thisSite = getNthSite(i);
	if (thisSite == (t_site *) NULL)
		{
		NotifyUser(-1, -1, "Error", "Unable to match site\n");
		}

	if (thisSite->display)
		{
		// highlight the current line
		if (!memcmp((void *) &config->curSite,
              	(void *) thisSite, sizeof(t_site)))
			width = 3;
		else
			width = 1;
	
		drawSite(thisSite, width);
		}
	}
}


/*----------------------------------------------------------------------------*/
// draw the given sampling Site
/*----------------------------------------------------------------------------*/
void drawSite(t_site * thisSite, int width)
{
int x, y;

if (thisSite == (t_site *) NULL)
	return;

pushGrEnv();

SetActiveWn(graphWn, 0);
SetClipArea();
PenSize( width, width );
PenMode(zNXORz);
PenPattern( 1 );
PenDash( 0 );
HideCursor();
PenColor(thisSite->type);

latLonToScreen(thisSite->pos.lat, thisSite->pos.lon, &x, &y);
draw_symbol(SQUARE_DOT, x, y);

PenSize( 0, 0 );
SetActiveWn(0, 0);
ShowCursor();
PenColor(BLACK);
popGrEnv();
}




/*----------------------------------------------------------------------------*/
// draw coast line
/*----------------------------------------------------------------------------*/
void drawCoastline(t_coast * theCoast)
{
register int i;
int  x, y;
register int penIsUp = FALSE;
t_setup * config = QuerySetup();

if ((theCoast == (t_coast *) NULL) || (!config->misc.showCoastline))
	return;

SetActiveWn(graphWn, 0);
SetClipArea();
PenSize( 1, 1 );
PenMode(zNXORz);
PenPattern( 1 );
PenDash( 0 );
HideCursor();
PenColor(BLUE);

SetActiveWn(graphWn, 0);
SetClipArea();

for(i=0; i< theCoast->numPts; i++)
	{
	if (theCoast->pts[i].lat == 0.0F)
		{
		penIsUp = TRUE;
		}
	else
		{
		latLonToScreen((double) theCoast->pts[i].lat, 
                    (double) theCoast->pts[i].lon, 
                     &x, &y);

		if (i == 0 || penIsUp)
			{
			MoveTo(x, y);
			}
		else
			{
			LineTo(x, y);
			}
		penIsUp = FALSE;
		}
	}

PenSize( 0, 0 );
PenMode(zNXORz);
SetActiveWn(0, 0);
ShowCursor();
PenColor(BLACK);
}



/*----------------------------------------------------------------------------*/
// Get Display Type for main panel
/*----------------------------------------------------------------------------*/
int doDisplayTypeFm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "disptype.frf";
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, NULL_SETUP_MSG, frf);
  return FALSE;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

// set up Type of diagnostic
switch(config->DisplayType)
	{
	case PLAN_DISPLAY:
		fm->ob_lst [ 2 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 4 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 5 ].flags &= ~IS_CHECKED;
		break;

	case LINE_DISPLAY:
		fm->ob_lst [ 3 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 2 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 4 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 5 ].flags &= ~IS_CHECKED;
		break;

	case POLAR_DISPLAY:
		fm->ob_lst [ 4 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 2 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 5 ].flags &= ~IS_CHECKED;
		break;

	case RANGE_REDUCTION_DISPLAY:
		fm->ob_lst [ 5 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 2 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
		fm->ob_lst [ 4 ].flags &= ~IS_CHECKED;
		break;
	}

OpenForm ( fm, ob->area.Xmin, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the flags
			config->DisplayType = get_flag_fm_group(fm, 1);
			RefreshRequired = TRUE;
			break;

		case CANCEL:
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}



/*----------------------------------------------------------------------------*/
// Get Display Type for side panel
/*----------------------------------------------------------------------------*/
int doSidePanelFm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "sidepanl.frf";
t_setup * config = QuerySetup();
int save;
char buf[256];

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, NULL_SETUP_MSG, frf);
  return FALSE;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

// set up Type of displa in side panel
switch(save = config->SidePanelType)
	{
	case LINE_INFO:
		fm->ob_lst [ 2 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
		break;

	case SITE_INFO:
		fm->ob_lst [ 3 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 2 ].flags &= ~IS_CHECKED;
		break;

	default:
		fm->ob_lst [ 2 ].flags |=  IS_CHECKED;
		fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
		break;
	}

OpenForm ( fm, ob->area.Xmin, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the flags
			config->SidePanelType = get_flag_fm_group(fm, 1);
			if (save != config->SidePanelType)
				{
				// remove current side panel display
				CloseForm(navDispFm);
				DestroyFRFForm ( navDispFm );
				navDispFm = (fm_desc *) NULL;
				
				if ( ( navDispFm = loadSidePanelFRF() ) == (fm_desc *) NULL ) 
					{
					Puke("Can't Load navDispFm");
					}

				// Open new Side panel display
				if (!(retv = OpenForm( navDispFm, navDispX, navDispY)))
					{
					sprintf(buf, "%s %s at line %d retv = %d navDispX = %d navDispY = %d\n", CANT_OPEN_MSG, __FILE__, __LINE__, retv, navDispX, navDispY);
					Puke(buf);
					}
				updateSensorName();
				updateDiskFree(1);
				}
			RefreshRequired = TRUE;
			break;

		case CANCEL:
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}



/*----------------------------------------------------------------------------*/
// Open if necessary and update the range reduction form
/*----------------------------------------------------------------------------*/
int updateRangeForm(t_nav * nav)
{
int i, k;
ftag * namTag;
ftag * idTag;
ftag * rngTag;
ftag * azmTag;
ftag * residTag;
ftag * iterTag;
ftag * timTag;
ftag * latTag;
ftag * lonTag;
ftag * deltaTag;
char buf[80];
double deg;
t_observations * obs = &nav->msmts.rng.obs;
double distance, fazm, bazm;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid for arclength comps

if (nav == (t_nav *) NULL)
	return NULL_OBSERVATIONS;

if (obs == (t_observations *) NULL)
	return NULL_OBSERVATIONS;

// Load the FRF file if necessary
if (rangeFm == (fm_desc *) NULL)
	{
	if (openRangeForm() != SUCCESS)
		{
  	return BAD_FORM_LOAD;
		}
	}

// Clear the old values
for (i=0; i < (rangeFm->ob_cnt - 1) / 5; i++)
	{
	namTag   = (ftag *) rangeFm->ob_lst[5*i+0].tag;
	strset(namTag->fld, '\0');

	idTag    = (ftag *) rangeFm->ob_lst[5*i+1].tag;
	strset(idTag->fld, '\0');

	rngTag   = (ftag *) rangeFm->ob_lst[5*i+2].tag;
	strset(rngTag->fld, '\0');

	residTag = (ftag *) rangeFm->ob_lst[5*i+3].tag;
	strset(residTag->fld, '\0');

	azmTag   = (ftag *) rangeFm->ob_lst[5*i+4].tag;
	strset(azmTag->fld, '\0');

	// redraw the fields
	for (k=0; k<5; k++)
		{
		DrawTaggedOb(&rangeFm->ob_lst[5 * i + k]);
		}
	}

// cycle through the elements in the form
for (i=0; i<obs->numObs && i < (rangeFm->ob_cnt / 5); i++)
	{
	namTag   = (ftag *) rangeFm->ob_lst[5*i+0].tag;
	strcpy(namTag->fld, obs->stations[i]->name);

	idTag    = (ftag *) rangeFm->ob_lst[5*i+1].tag;
	strcpy(idTag->fld, itoa(obs->stations[i]->id, buf, 10));

	rngTag   = (ftag *) rangeFm->ob_lst[5*i+2].tag;
	sprintf(rngTag->fld, "%.0lf", obs->ranges[i]);

	residTag = (ftag *) rangeFm->ob_lst[5*i+3].tag;
	sprintf(residTag->fld, "%.1lf", obs->residuals[i]);

	azmTag   = (ftag *) rangeFm->ob_lst[5*i+4].tag;
	deg = RAD_TO_DEG * obs->bearings[i];
	if (deg < 0.0)
		deg += 360.0;
	sprintf(azmTag->fld, "%.1lf", deg);

	// redraw the fields
	for (k=0; k<5; k++)
		{
		DrawTaggedOb(&rangeFm->ob_lst[5 * i + k]);
		}
	}

// show the iterations
iterTag = (ftag *) rangeFm->ob_lst[rangeFm->ob_cnt - 5].tag;
strcpy(iterTag->fld, itoa(obs->iterations, buf, 10));
DrawTaggedOb(&rangeFm->ob_lst[rangeFm->ob_cnt - 5]);

// show the time of fix
timTag = (ftag *) rangeFm->ob_lst[rangeFm->ob_cnt - 4].tag;
strcpy(timTag->fld, getTimeStr((double) nav->tim));
DrawTaggedOb(&rangeFm->ob_lst[rangeFm->ob_cnt - 4]);

// show the lat and lon
latTag = (ftag *) rangeFm->ob_lst[rangeFm->ob_cnt - 3].tag;
lonTag = (ftag *) rangeFm->ob_lst[rangeFm->ob_cnt - 2].tag;
PositionString(lonTag->fld, latTag->fld, buf, nav->lon, nav->lat, 0.0, 0);
DrawTaggedOb(&rangeFm->ob_lst[rangeFm->ob_cnt - 3]);
DrawTaggedOb(&rangeFm->ob_lst[rangeFm->ob_cnt - 2]);

// compute distance to primary nav system
if (validNav(nav) == SUCCESS)
	{
	arclen(nav->lat, nav->lon,
         curNav->lat, curNav->lon,
         &distance,
         &fazm,
         &bazm,
         ellipse->a,
         ellipse->b);

	deltaTag = (ftag *) rangeFm->ob_lst[rangeFm->ob_cnt - 1].tag;
	sprintf(deltaTag->fld, 
          "%.1lfm at %3.1lf%c", 
          distance, fazm * RAD_TO_DEG, 248);
	DrawTaggedOb(&rangeFm->ob_lst[rangeFm->ob_cnt - 1]);
	}

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// load and open the range form
/*----------------------------------------------------------------------------*/
int openRangeForm(void)
{
int retv;
char * frf = "range.frf";
char buf[128];

// Load the FRF file 
if ((rangeFm = loadFRF ( frf ) ) == (fm_desc *) NULL ) 
	{
  return BAD_FORM_OPEN;
	}

if (!(retv = OpenForm( rangeFm, graphWnRect.Xmin, graphWnRect.Ymin)))
	{
	sprintf(buf, "%s %s at line %d retv = %d x = %d y = %d\n", 
         CANT_OPEN_MSG, 
         __FILE__, 
         __LINE__, 
         retv, 
         graphWnRect.Xmin, 
         graphWnRect.Ymin);
	Puke(buf);
	}

return SUCCESS;
}




/*----------------------------------------------------------------------------*/
// load and open the range form
/*----------------------------------------------------------------------------*/
int killRangeForm(void)
{
if (rangeFm == (fm_desc *) NULL)
	return -1;

CloseForm   ( rangeFm );
DestroyFRFForm ( rangeFm );
rangeFm = (fm_desc *) NULL;
return SUCCESS;
}




/*----------------------------------------------------------------------------*/
// Draw the passed track history
/*----------------------------------------------------------------------------*/
void drawHistory(t_history * p)
{
int i;
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();

if (p == (t_history *) NULL)
	{
	scrollMsg("Null history\n");
	return;
	}

// Set up graphics
SetActiveWn(graphWn, 0);
pushGrEnv();

SetClipArea();
HideCursor();

for (i=0; i<p->numPts; i++)
	{
	// get screen coords
	drawMarker(p->fix[i].pos.lat, 
             p->fix[i].pos.lon, 
             (int) config->misc.fixSymbol,
             zREPz,
             curNav->id);
             
	}

PenSize( 0, 0 );
popGrEnv();

SetActiveWn(lastWn, 0);
ShowCursor();
}




/*----------------------------------------------------------------------------*/
// Draw all trail lines 
/*----------------------------------------------------------------------------*/
void drawTrailLines(void)
{
int i;
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();
t_nav * pNav;
int retv;

// Always draw the curNav sensor
drawTrailLine(curNav->id);

// loop through the active sensor list and draw the other sensors
for (i=0; i<MAX_COMPORTS; i++)
	{
	if (config->displaySensors[i])
		{
		pNav = queryNavPtr(i);
		if ((retv = validNav(pNav)) == SUCCESS)
			{
			drawTrailLine(pNav->id);
			}
		else
			{
			scrollMsg("Can't draw\n  ");
			scrollMsg(com_struct[i].sensor_name);
			}
		}
	}
}




/*----------------------------------------------------------------------------*/
// Draw the ships current track position
/*----------------------------------------------------------------------------*/
void drawTrailLine(int color)
{
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();
t_pos * vrpPos;

// Set up graphics
SetActiveWn(graphWn, 0);
SetClipArea();

pushGrEnv();

HideCursor();
vrpPos = queryVRP();
drawMarker(vrpPos->lat, vrpPos->lon, config->misc.fixSymbol,
           zREPz,
           color);
	
popGrEnv();

SetActiveWn(lastWn, 0);
ShowCursor();
}




/*----------------------------------------------------------------------------*/
// Get Filter Constant
/*----------------------------------------------------------------------------*/
int doFilterForm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "filter.frf";
t_setup * config = QuerySetup();
ftag * filterTimeTag;
ftag * filterSmoothTag;
double d;
int i;

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, NULL_SETUP_MSG, frf);
  return FALSE;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

// set up form
filterTimeTag     = (ftag *) fm->ob_lst[2].tag;
filterSmoothTag   = (ftag *) fm->ob_lst[3].tag;
sprintf(filterTimeTag->fld, "%.0lf",  config->filter.time);
sprintf(filterSmoothTag->fld, "%d", config->filter.weight);

OpenForm ( fm, ob->area.Xmin, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the fields
			d = atof(filterTimeTag->fld);
			if (d <= 0.0)
				 d = 1.0;
			config->filter.time = d;

			i = atoi(filterSmoothTag->fld);

			if (i <= 0)
				 i = 1;

			if (i > 100)
				 i = 100;

			config->filter.weight = i;
			break;

		case CANCEL:
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}



/*----------------------------------------------------------------------------*/
// Get the screen coordinates of the ship for LINE_DISPLAY mode
/*----------------------------------------------------------------------------*/
int getLineOffsets(int * x, int * y)
{
t_track * pTrk = queryTrackInfo();

if (pTrk == (t_track *) NULL)
	{
	scrollMsg(NULL_TRACK_MSG);
	return NULL_TRACK;
	}

// x coordinate will be related to cross track distance and screen center
*x = (int) (pTrk->cross * graf.xres) + graf.centerX;

// y coordinate will be related to along track distance and screen center
*y = (int) (pTrk->along * graf.xres) + graf.centerX;
}




/*----------------------------------------------------------------------------*/
// Set Clip area
/*----------------------------------------------------------------------------*/
void SetClipArea(void)
{
SetActiveWn(graphWn, 0);
ClipRect(&clipRect);
}


/*----------------------------------------------------------------------------*/
// Clear the graphics drawing area
/*----------------------------------------------------------------------------*/
void ClearGraphicsArea(void)
{
int lastWn = QryActiveWn();

SetActiveWn(graphWn, 0);
ClipRect(&clipRect);
EraseRect(&clipRect);
SetActiveWn(lastWn, 0);
ShipIsDrawn = FALSE;
FishIsDrawn = FALSE;
}

/*----------------------------------------------------------------------------*/
// Draw the North Indicator
/*----------------------------------------------------------------------------*/
void textAnnot(char * label, int x, int y, int xSize, int ySize)
{
// int graphicsAngle = (int) ((90.0 - theLine->bearing * RAD_TO_DEG) * 10.0);
int lastWn = QryActiveWn();

// see if they sent a valid string
if (label == (char *) NULL)
	return;

SetActiveWn(graphWn, 0);
SetClipArea();

pushGrEnv();
LoadFont( "ROMANSIM.FNT" );

// TextPath(graphicsAngle);
PenColor(0);
TextMode(zREPz);
TextSize( xSize, ySize );
TextFace( cNormal );
PenMode(zREPz);

HideCursor();
MoveTo(x, y);
DrawText(label, 0, strlen(label));
ShowCursor();

TextSize( 0, 0 );
TextPath(0);
TextSlant( 0 );
TextFace( cProportional );

SetMnFont();
popGrEnv();
SetActiveWn(lastWn, 0);
}



/*----------------------------------------------------------------------------*/
// Draw the Fish
/*----------------------------------------------------------------------------*/
void drawFish(void)
{
static int last_x = 0, last_y = 0;
static double last_hdg = 0.0;
int x, y;
double hdg;
t_nav fishNav;
int retv;
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();

// get the data for the fish
if ((retv = queryFishNav(&fishNav)) != SUCCESS)
	{
	// scrollMsg("No Fish Nav\n");
	return;
	}

// Only plot the EG & G
if (fishNav.msmts.usbl.transponder_id != 1)
	{
	scrollMsg("Wrong Fish ID\n");
	return;
	}

SetClipArea();
pushGrEnv();

// PenMode(zNXORz);
// PenSize( 3, 3 );
// PenPattern( 1 );
// PenDash( 0 );
// HideCursor();

latLonToScreen(fishNav.lat, fishNav.lon, &x, &y);

hdg = curNav->crs;          // use the course made good for fish heading
eraseFish();

drawRotatableObject(&Fish, hdg, x, y, config->misc.shipScaling);
fishLat = fishNav.lat;
fishLon = fishNav.lon;
fishHdg = hdg;
	
last_x    = x;
last_y    = y;
last_hdg  = hdg;
FishIsDrawn = TRUE;

ShowCursor();
popGrEnv();
SetActiveWn(lastWn, 0);
}



/*----------------------------------------------------------------------------*/
// Erase the Fish
/*----------------------------------------------------------------------------*/
void eraseFish(void)
{
int x, y;
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();

if (!FishIsDrawn)
	return;

if (fishLat != 0.0 && fishLon != 0.0)
	{
	SetClipArea();
	pushGrEnv();
	PenMode(zNXORz);
	// PenSize( 3, 3 );
	// PenPattern( 1 );
	// PenDash( 0 );
	// HideCursor();
	if (!latLonToScreen(fishLat, fishLon, &x, &y))
		{
		popGrEnv();
		return;                       // don't try to erase if it's not on screen
		}
	drawRotatableObject(&Fish, fishHdg, x, y, config->misc.shipScaling);
	// ShowCursor();
	SetActiveWn(lastWn, 0);
	popGrEnv();
	}
}

/*----------------------------------------------------------------------------*/
// get a pointer to the nav struct of the fish & see if its active
/*----------------------------------------------------------------------------*/
int queryFishNav(t_nav * destNav)
{
int retv;
// request the nav structure
t_nav * pNav = queryNavByType(TRACKPOINT_II);
if (pNav == (t_nav *) NULL)
	{
	return NO_FISH_NAV;
	}
else
	{
	memcpy((void *) destNav, (void *) pNav, sizeof(t_nav));
	}
retv = validNav(pNav);
return retv;
}


/*----------------------------------------------------------------------------*/
// draw a marker
/*----------------------------------------------------------------------------*/
void drawMarker(double lat, double lon, 
                int symbol,
                int penMode, 
                int penColor)
{
int x, y;
int lastWn = QryActiveWn();

SetActiveWn(graphWn, 0);
SetClipArea();
PenSize( 1, 1 );
PenMode(penMode);
PenPattern( 1 );
PenDash( 0 );
HideCursor();
PenColor(penColor);

latLonToScreen(lat, lon, &x, &y);
draw_symbol(symbol, x, y);

PenSize( 0, 0 );
SetActiveWn(0, 0);
ShowCursor();
PenColor(BLACK);
SetActiveWn(lastWn, 0);
}


/*----------------------------------------------------------------------------*/
// Set Miscellaneous display options
/*----------------------------------------------------------------------------*/
int doDisplayMisc ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "dispMisc.frf";
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	NotifyUser(-1, -1, NULL_SETUP_MSG, frf);
  return FALSE;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

if (config->misc.shipCentering)
	{
	fm->ob_lst [ 2 ].flags |=  IS_CHECKED;
	}
else
	{
	fm->ob_lst [ 2 ].flags &= ~IS_CHECKED;
	}

if (config->misc.shipScaling)
	{
	fm->ob_lst [ 3 ].flags |=  IS_CHECKED;
	}
else
	{
	fm->ob_lst [ 3 ].flags &= ~IS_CHECKED;
	}


if (config->misc.showCoastline)
	{
	fm->ob_lst [ 4 ].flags |=  IS_CHECKED;
	}
else
	{
	fm->ob_lst [ 4 ].flags &= ~IS_CHECKED;
	}

// Clear the symbology flags
fm->ob_lst [ 5 ].flags &= ~IS_CHECKED;
fm->ob_lst [ 6 ].flags &= ~IS_CHECKED;
fm->ob_lst [ 7 ].flags &= ~IS_CHECKED;

// Set symbology flgs
switch (config->misc.fixSymbol)
	{
	case DOT:
		fm->ob_lst [ 5 ].flags |= IS_CHECKED;
		break;

	case CROSS:
		fm->ob_lst [ 6 ].flags |= IS_CHECKED;
		break;

	case THE_X:
		fm->ob_lst [ 7 ].flags |= IS_CHECKED;
		break;
	}


OpenForm ( fm, ob->area.Xmin, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// ship Centering
			if (fm->ob_lst [ 2 ].flags & IS_CHECKED)
				{
				config->misc.shipCentering = TRUE;
				}
			else
				{
				config->misc.shipCentering = FALSE;
				}

			// ship scaling
			if (fm->ob_lst [ 3 ].flags & IS_CHECKED)
				{
				config->misc.shipScaling = TRUE;
				}
			else
				{
				config->misc.shipScaling = FALSE;
				}

			// coastline enabled
			if (fm->ob_lst [ 4 ].flags & IS_CHECKED)
				{
				config->misc.showCoastline = TRUE;
				}
			else
				{
				config->misc.showCoastline = FALSE;
				}

			config->misc.fixSymbol = get_flag_fm_group(fm, 2);

			RefreshRequired = TRUE;	
			break;

		case CANCEL:
			break;

		default : 
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}


/*----------------------------------------------------------------------------*/
// return the appropriate side panel frf filename
/*----------------------------------------------------------------------------*/
fm_desc * loadSidePanelFRF(void)
{
t_setup * config = QuerySetup();
char frf_file [30];

if (config->SidePanelType == LINE_INFO)
	strcpy(frf_file,  "linedisp.frf");
else
	strcpy(frf_file,  "sitedisp.frf");

navDispFm = loadFRF(frf_file);

return navDispFm;
}

/*----------------------------------------------------------------------------*/
// return a pointer to the clipRect
/*----------------------------------------------------------------------------*/
rect * queryPrintRect(void)
{
return (rect *) &clipRect;
}



/*----------------------------------------------------------------------------*/
// Initialize the Polar display.
// Compute size of scale that will allow the ship & the site to be in display.
// Put the Site Symbol in the Center of the Display.
// Draw North Indicator
/*----------------------------------------------------------------------------*/
void initPolarDisplay(void)
{
int lastWn = QryActiveWn();
t_setup * config = QuerySetup();
long   lRadius;

setCenter(config->curSite.pos.lat, config->curSite.pos.lon);

// compute radius of smallest range ring
lRadius = graf.scale / 2L;                       // exact size of rings
lRadius -= lRadius % 10L;                        // keep it on 10 m boundaries
lRadius = max(lRadius, 50L);                     // nothing smaller than 50m
drawRangeRings(lRadius);

// Draw the site in the screen center
drawSite(&config->curSite, 5);

SetActiveWn(lastWn, 0);
}



/*----------------------------------------------------------------------------*/
// Compute size of scale that will allow the ship & the site to be in display.
/*----------------------------------------------------------------------------*/
void drawRangeRings(long radius)
{
rect r;
int x = 4 + graf.centerX;
int y = graf.centerY;
int x1, y1;                                 // lower left of ellipse rect
int x2, y2;                                 // upper right of ellipse rect
int i;
int lastWn = QryActiveWn();
char buf[80];
double thisRadius;
double theStep;
double utmX, utmY;

SetClipArea();
pushGrEnv();

PenMode(zNXORz);
PenSize( 3, 3 );
PenPattern( 1 );
PenDash( 0 );
HideCursor();

theStep    = (double) (radius) / (double) NUM_RANGE_RINGS;
thisRadius = (double) radius;
for (i=0; i<NUM_RANGE_RINGS; i++)
	{
	utmX = graf.center_utm.x - thisRadius;
	utmY = graf.center_utm.y + thisRadius;

	// get the pixel locations for lower left x & y
	utmToScreen(utmX, utmY, &x1, &y1);

	utmX = graf.center_utm.x + thisRadius;
	utmY = graf.center_utm.y - thisRadius;

	// get the pixel locations for upper right x & y
	utmToScreen(utmX, utmY, &x2, &y2);

	// build the rectangle
	SetRect(&r, x1, y1, x2, y2);

	// Draw it
	FrameOval(&r);

	// Label it
	sprintf(buf, "%.0lf", thisRadius);
	PenSize( 1, 1 );
	textAnnot(buf, r.Xmin + 2, r.Ymin + (r.Ymax - r.Ymin) / 2, 7, 12);
	PenSize( 3, 3 );

	// set up for next pass
	thisRadius -= theStep;
	}

popGrEnv();
SetActiveWn(lastWn, 0);
}

/*----------------------------------------------------------------------------*/
// dump a rectangle
/*----------------------------------------------------------------------------*/
char * dumpRect(rect * r)
{
int j = 0;
static char buf[256];

if (r == (rect *) NULL)
	{
	j += sprintf(buf, "Null Rectangle\n");
	}
else
	{
	j += sprintf(buf+j, "X1 = %d Y1 = %d\n", r->Xmin, r->Ymin);
	j += sprintf(buf+j, "X2 = %d Y2 = %d\n", r->Xmax, r->Ymax);
	j += sprintf(buf+j, "width  = %d\n", r->Xmax - r->Xmin);
	j += sprintf(buf+j, "height = %d\n", r->Ymax - r->Ymin);
	}
return (char *) buf;
}


