/*----------------------------------------------------------------------------*/
// File: Nav.c   USGS Navigation
// Author: John Gann
// Start Date: 5 Dec 1991
/*----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <fcntl.h>
#include <io.h>
#include <bios.h>

#include <GRconst.h>
#include <GRports.h>
#include <GRextrn.h>
#include <menuet.h>

#include "navdefs.h"
#include "navtypes.h"
#include "nav.h"
#include "lines.h"
#include "sensor.h"
#include "serial.h"
#include "fileio.h"
#include "nav_util.h"
#include "graph.h"
#include "geo_util.h"
#include "stations.h"
#include "dispasci.h"
#include "offsets.h"
#include "mnutils.h"
#include "video.h"
#include "config.h"
#include "tracking.h"
#include "coast.h"
#include "logbook.h"
#include "depthcor.h"
#include "sites.h"
#include "contact.h"
#include "relay.h"
#include "plot.h"
#include "shapes.h"
#include "ec_comm.h"
#include "criterr.h"
#include "printer.h"
#include "sysinfo.h"


// global data
char _Debug = FALSE;
char _SmartPortInstalled = FALSE;
fm_desc * mainMenu;
char DisplayingSomething = FALSE;
char RefreshRequired = FALSE;
t_nav * curNav;
t_com * curCom;
int (*usrKeyTask)(ob_desc * ob);                       // used by DisplayAscii
int depthAlarmFlag = 0;                                     // used with PDD

// External data
extern void process_inspection_port_key( void );	
extern t_com com_struct[];
extern int videoWn;
extern fm_desc * navDispFm;
extern int postPlotActive;

static char * NULL_COM_MSG = "Null COM pointer";
static char * NULL_NAV_MSG = "Null NAV pointer";
static t_history shipHistory;                   // ship's track
static char * NULL_SETUP_MSG  = "Null Configuration";


/*----------------------------------------------------------------------------*/
/* void init(int argc, char *argv[])                                          */
/*----------------------------------------------------------------------------*/
int init_nav(int argc, char *argv[])
{
t_setup * config = QuerySetup();
int msType;
int msIRQ;

// initialize the printer for streamed IO
setmode ( fileno ( stdprn ), O_TEXT );

/* Set time zone variable to GMT0 */
if (putenv("TZ=GMT0"))
	{
	printf("\nCan't Set environment variable TZ");
	}
else
	{
	tzset();
	}

/* Initialize Menuet stuff */
InitApp( argc, argv );

if (QueryColors() <= 1)
  {
  SetAttrTblColors ( 0, 15 );
  }
else
  {
  SetAttrTblColors ( config->fg, config->bg );
  }


/* Try to Load the MRF FILE */
if ( ( mainMenu = LoadMRF ( "nav.mrf" ) ) == NULL ) 
	{
	Puke("Error Loading MRF\n");
	}

OpenApp( NAV_VERSION , mainMenu, NULL);
OpenMsgQueue( 1 );

/* Set the Esc Key up to cause immediate exit */
SetEscape(ESC_ESCAPE|ESC_MOUSE);

/* Set the number of timer ticks between to something really fast  */
SetDblClickInt (4);

/* Turn off Ghosting */
SetWnGhosting(0);

// Install Crical Error Handler
InstallCEH();

// read in the coms file 
if (!read_com_file(COM_FILE_NAME))
	{
	NotifyUser(-1, -1, COM_FILE_NAME, "Error reading COMS file");
	Puke("Error reading COMS file\n");
	}

// read in the nav setup file 
if (!read_nav(NAV_FILE_NAME))
	{
	NotifyUser(-1, -1, NAV_FILE_NAME, "Error reading NAV file");
	Puke("Error reading NAV file\n");
	}

// read in the nav stations
if (read_stations() <= FALSE)
	{
	NotifyUser(-1, -1, STATION_FILE_NAME, "Error reading Stations file");
	}

// read in the coastline
if (read_coastline() <= FALSE)
	{
	NotifyUser(-1, -1, "Yo", "Error reading Coastline file");
	}

// read in the antenna offsets
if (read_offsets(OFFSET_FILE_NAME) <= FALSE)
	{
	NotifyUser(-1, -1, OFFSET_FILE_NAME, "Error reading offsets file");
	}

// create the nav display window 
buildNavDisplay();

// Initialize File IO
if (initFileIO() != SUCCESS)
	{
	NotifyUser(-1, -1, "Yo", "Nav Data Logging is NOT Enabled");
	}

// if SmartPort not installed then disable mouse use if its a serial mouse
if (!_SmartPortInstalled && (qryMouse(&msType, &msIRQ) == SERIAL_MOUSE))
	StopMouse();

// Initialize serial IO
initAllPorts();

// init contact closure
initRelayIO();

// init printer as LPT1
openPrinter("LPT1");

return TRUE;
}







/*------------------------------------------------------------------*/
/* Give Not Yet Implemented message                                 */
/*------------------------------------------------------------------*/
int not_implemented(ob_desc * ob)
{
NotifyUser(-1, -1, ob->label, "Yo, this feature not yet implemented");
return TRUE;
}




/*------------------------------------------------------------------*/
// Return a pointer to the main menu
/*------------------------------------------------------------------*/
fm_desc *  QueryMenu(void)
{
return (fm_desc *) mainMenu;
}

/*------------------------------------------------------------------*/
// Process command line flags 
/*------------------------------------------------------------------*/
void process_cmd_line_flags(int argc, char ** argv)
{
int i;

for (i=1; i<argc; i++)
	{
	if ( (!strnicmp(argv[i], "-d", 2)) ||
       (!strnicmp(argv[i], "/d", 2)) )
		{
		_Debug = TRUE;
		}
	}
}




/*------------------------------------------------------------------*/
// the Exit function
/*------------------------------------------------------------------*/
int doExit(ob_desc * ob)
{
Puke("Terminated by Operator");
return TRUE;
}



/*------------------------------------------------------------------*/
/* Assign sub menus and tasks to main menu items										*/
/*------------------------------------------------------------------*/
void InitNavMenus(void)
{
fm_desc * thisFm;
t_setup * config = QuerySetup();

/* File SubMenu */
thisFm = (fm_desc *) mainMenu->ob_lst[0].tag;            

thisFm->ob_lst[0].task = doFileLoggingFm;          // Logging form
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doPrinterFm;              // Printer setup
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doChangeDiskette;         // Change the diskette
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doExit;                   // Exit form
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

/* Edit Sensor Sub Sub Menu */
thisFm = (fm_desc *) thisFm->ob_lst[1].tag;        // Line Sub Sub Menu
thisFm->ob_lst[0].task = doSensorForm;          // Sensor selection
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

/* Edit Sub Menu */
thisFm = (fm_desc *) mainMenu->ob_lst[1].tag;            
thisFm = (fm_desc *) thisFm->ob_lst[0].tag;            

thisFm->ob_lst[0].task = doActiveSensor;          // Select active Sensor
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doSensorForm;            // Port Assignment
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doGyroSelection;         // Pick a Gyro for Nav
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doStationForm;           // Nav Stations
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[4].task = doFilterForm;            // Filter form
thisFm->ob_lst[4].flags |= IS_EXECUTABLE;

thisFm->ob_lst[5].task = doCalibrateFm;            // Range Calibration form
thisFm->ob_lst[5].flags |= IS_EXECUTABLE;

thisFm = (fm_desc *) mainMenu->ob_lst[1].tag;     // Geodesy Sub Menu
thisFm = (fm_desc *) thisFm->ob_lst[2].tag;            

thisFm->ob_lst[0].task = doUtmZone;                // Default UTM Zone
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm = (fm_desc *) mainMenu->ob_lst[1].tag;     // Vessel sub sub menu
thisFm = (fm_desc *) thisFm->ob_lst[3].tag;            

thisFm->ob_lst[0].task = doShapeListFm;           // Vesel Outline Shape
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doOffsetsFm;            // Offsets form
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doVrpFm;                // VRP form
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm = (fm_desc *) mainMenu->ob_lst[1].tag;     // Sampling Sites 
thisFm = (fm_desc *) thisFm->ob_lst[4].tag;       // Sampling Sites Menu

thisFm->ob_lst[0].task = doAddSite;              // Add
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doEditSite;             // Edit
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doDeleteSites;          // Delete
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doSelectSite;           // Select
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[4].task = doListSites;            // List
thisFm->ob_lst[4].flags |= IS_EXECUTABLE;

// Edit Lines Sub Sub Menu
thisFm = (fm_desc *) mainMenu->ob_lst[1].tag;     // Vessel sub sub menu
thisFm = (fm_desc *) thisFm->ob_lst[1].tag;        // Line Sub Sub Menu

thisFm->ob_lst[0].task = doAddLine;                // Add Line
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doEditLine;               // Edit Line
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doDeleteLine;             // Delete Line
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doListLines;              // list Line
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[4].task = doSelectLine;             // Select Line
thisFm->ob_lst[4].flags |= IS_EXECUTABLE;

thisFm->ob_lst[5].task = doLineGeneratorFm;          // Build a grid of lines
thisFm->ob_lst[5].flags |= IS_EXECUTABLE;

/* Utilities Sub Menu */
thisFm = (fm_desc *) mainMenu->ob_lst[2].tag;            

thisFm->ob_lst[0].task = doRangeBearingFm;          // Range Bearing
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doInverseFm;            // Inverse
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doUTMtoLatLon;          // UTM to LL
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doLogBook;              // Logbook
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[4].task = doPlotFm;               // Post Plot Navigation
thisFm->ob_lst[4].flags |= IS_EXECUTABLE;

thisFm->ob_lst[5].task = viewVideoFile;          // Look at any video file
thisFm->ob_lst[5].flags |= IS_EXECUTABLE;

thisFm->ob_lst[6].task = doViewAsciiFile;        // Look at any ascii file
thisFm->ob_lst[6].flags |= IS_EXECUTABLE;

// Display
thisFm = (fm_desc *) mainMenu->ob_lst[3].tag;            

thisFm->ob_lst[0].task = doCoastLineFm;          // Coastlines
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doDisplayLines;          // Lines
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doVideo;                 // Video
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

if (!config->videoEnabled)
	{
	// GhostObject(&thisFm->ob_lst[2]);                // ghost the video selection
	thisFm->ob_lst[2].flags &= ~IS_SELECTABLE;
	}

thisFm->ob_lst[3].task = doTrackingFm;            // Tracking
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[5].task = doSetCenter;             // Set Center
thisFm->ob_lst[5].flags |= IS_EXECUTABLE;

thisFm->ob_lst[6].task = doDisplayTypeFm;          // Pick a main panel Type
thisFm->ob_lst[6].flags |= IS_EXECUTABLE;

thisFm->ob_lst[7].task = doSidePanelFm;            // Pick a side panel Type
thisFm->ob_lst[7].flags |= IS_EXECUTABLE;

thisFm->ob_lst[8].task = doDisplayMisc;            // Misc. Display Options
thisFm->ob_lst[8].flags |= IS_EXECUTABLE;

thisFm = (fm_desc *) thisFm->ob_lst[4].tag;        // Display Set Scale Submenu

thisFm->ob_lst[0].task = doPickScale;              // Choose Scale
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doRubberRect;             // Rubber Rect Scale
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

// Config
thisFm = (fm_desc *) mainMenu->ob_lst[4].tag;            

thisFm->ob_lst[0].task = doCoordSysFm;            // Coordinate System
thisFm->ob_lst[0].flags |= IS_EXECUTABLE;

thisFm->ob_lst[1].task = doUnitsFm;               // Units 
thisFm->ob_lst[1].flags |= IS_EXECUTABLE;

thisFm->ob_lst[2].task = doDepthFm;              // Depth Adjustments
thisFm->ob_lst[2].flags |= IS_EXECUTABLE;

thisFm->ob_lst[3].task = doBridgeCfgFm;            // Bridge Terminal Setup
thisFm->ob_lst[3].flags |= IS_EXECUTABLE;

thisFm->ob_lst[4].task = doRelayFm;               // Contact Closure Setup
thisFm->ob_lst[4].flags |= IS_EXECUTABLE;

return;
}




/*------------------------------------------------------------------*/
// the main function
/*------------------------------------------------------------------*/
void main(int argc, char ** argv)
{
unsigned int ioPort;

process_cmd_line_flags(argc, argv);

if (_SmartPortInstalled = (char) SmartPortInstalled(FALSE,  &ioPort))
	{
	if (_Debug)
		{
		printf("SmartPort is installed at IO address %04X\n", ioPort);
		getch();
		}
	}
else
	{
	printf("Warning: SmartPort is Not installed\n");
	// getch();
	if (ecInit() == SUCCESS)
		{
		setIOTasks(ecGetChar, ecPutChar, ecInputQueue);
		}
	}

init_nav(argc, argv);
InitNavMenus();

// Set the key Task
SetUserKeyTask(navKeyTask);
usrKeyTask = navKeyTask;              // required forDisplayAscii

// Set the background task
SetUserEchoTask(BackGroundTask);

logMsg("StartUp Nav");

while (1)
	ProcessMenu(mainMenu);
}





/*----------------------------------------------------------------------------*/
// Terminate from anywhere
/*----------------------------------------------------------------------------*/
void Puke(char * msg)
{
logMsg(msg);                      // Put Msg in logbook
closeLogFiles();                  // Close all logfiles
RemoveCEH();                      // remove Dos critical error handler
termVideo();                      // close the video board
clearAllRelays();                 // Make sure the relays don't get stuck open
disable_serial_ports();           // disable serial interrupts
write_video(VIDEO_FILE_NAME);     // update the video file settings
write_setup();                    // update configuration
write_nav(NAV_FILE_NAME);         // update last position speed, etc
write_line_file();                // update the line list
write_offsets(OFFSET_FILE_NAME);  // update the vessel offset file
write_station_file();             // update the station list
write_com_file(COM_FILE_NAME);    // update the serial port assignments
write_graf(GRAF_FILE_NAME);       // update the graphics configuration
CloseMsgQueue();                  // Shut down menuet and Metagraphics
TermApp(msg);                     // Close the application
}



/*----------------------------------------------------------------------------*/
// Backgound Task
/*----------------------------------------------------------------------------*/
void BackGroundTask (int x, int y)
{
static time_t lastTime = 0L;
time_t now = time(NULL);

// if ((_SmartPortInstalled) && (curCom != (t_com *) NULL))
if (curCom != (t_com *) NULL)
	{
	readAllPorts();
	recordAsciiNav();
	}

if (now - lastTime)
	{
	ShowCursor();
	sendAllPorts();
	updateShipHistory(&shipHistory);
	doNavComps(curNav, &shipHistory);
	computeVRP(curNav);
	displayNav();
	printNav();
	doGraphics();
	doRelays();
	if (postPlotActive)
		{
		doPostPlotting();
		}
	lastTime = now;
	}

updateCursorCoords(x, y, FALSE);

if (RefreshRequired)
	{
	if (refreshGraphics() == SUCCESS)
		{
		RefreshRequired = FALSE;
		updateCursorCoords(x, y, TRUE);           // force coordinates to be updated
		}
	}
}


/*----------------------------------------------------------------------------*/
// function to dump all msgs to scroll window
/*----------------------------------------------------------------------------*/
int showAllPorts (void)
{
int i;

for (i=0; i<MAX_COMPORTS; i++)
	{
	if (com_struct[i].enabled)
		{
		if (com_struct[i].buf[0] != '\0')
			{
			scrollMsg(com_struct[i].buf);
			}
		}
	}
return i;
}


/*----------------------------------------------------------------------------*/
// Update the line info side panel
/*----------------------------------------------------------------------------*/
void updateLineInfo(void)
{
static int lastSensorIdx = -1L;
static char lineName[30];
ftag * latTag;
ftag * lonTag;
ftag * timTag;
ftag * spdTag;
ftag * crsTag;
ftag * gyrTag;
ftag * navTag;
ftag * croTag;
ftag * aloTag;
ftag * endTag;
ftag * linTag;
ftag * brgTag;
ftag * rngTag;
char buf[80];
int retv;
t_nav * gyroNav = (t_nav *) NULL;       // pointer to Gyro nav struct
t_track * track = (t_track *) NULL;     // pointer to track info struct

if ((DisplayingSomething) || navDispFm == (fm_desc *) NULL)
	{
	return;
	}
else
	{
	if ((retv = validNav(curNav)) != SUCCESS)
		return;
	}

timTag = (ftag *) navDispFm->ob_lst[0].tag;
latTag = (ftag *) navDispFm->ob_lst[1].tag;
lonTag = (ftag *) navDispFm->ob_lst[2].tag;
spdTag = (ftag *) navDispFm->ob_lst[3].tag;
crsTag = (ftag *) navDispFm->ob_lst[4].tag;
gyrTag = (ftag *) navDispFm->ob_lst[5].tag;
navTag = (ftag *) navDispFm->ob_lst[6].tag;
croTag = (ftag *) navDispFm->ob_lst[7].tag;
aloTag = (ftag *) navDispFm->ob_lst[8].tag;
endTag = (ftag *) navDispFm->ob_lst[9].tag;
linTag = (ftag *) navDispFm->ob_lst[10].tag;
rngTag = (ftag *) navDispFm->ob_lst[11].tag;
brgTag = (ftag *) navDispFm->ob_lst[12].tag;
	
PositionString(lonTag->fld, latTag->fld, buf, 
               curNav->lon, curNav->lat, 0.0, 0);

strcpy(timTag->fld, getTimeStr(0.0));
sprintf(spdTag->fld, "%5.1lf", curNav->spd);
sprintf(crsTag->fld, "%5.1lf", curNav->crs * RAD_TO_DEG);

// get the active gyro if there is one
gyroNav = queryNavByType(QuerySetup()->Gyro);
if (gyroNav == (t_nav *) NULL)
	{
	// must not be a gyro connected so use the crs made good
	sprintf(gyrTag->fld, "%5.1lf", curNav->crs * RAD_TO_DEG);
	// scrollMsg(NULL_COM_MSG);
	}
else
	{
	curNav->hdg = gyroNav->hdg;
	sprintf(gyrTag->fld, "%5.1lf", (curNav->hdg = gyroNav->hdg) * RAD_TO_DEG);
	}

// get the track info stuff
track = queryTrackInfo();
if (track == (t_track *) NULL)
	{
	scrollMsg("Null track\n");
	return;
	}
else
	{
	sprintf(croTag->fld, "%6.0lf", track->cross);
	sprintf(aloTag->fld, "%6.0lf", track->along);
	sprintf(endTag->fld, "%6.0lf", track->toend);
	}

DrawTaggedOb(&navDispFm->ob_lst[0]);
DrawTaggedOb(&navDispFm->ob_lst[1]);
DrawTaggedOb(&navDispFm->ob_lst[2]);
DrawTaggedOb(&navDispFm->ob_lst[3]);
DrawTaggedOb(&navDispFm->ob_lst[4]);
DrawTaggedOb(&navDispFm->ob_lst[5]);

DrawTaggedOb(&navDispFm->ob_lst[7]);
DrawTaggedOb(&navDispFm->ob_lst[8]);
DrawTaggedOb(&navDispFm->ob_lst[9]);

// Use the sensor index to get the associated port
if (curCom == (t_com *) NULL)
	{
	// NotifyUser(-1, -1, __FILE__, NULL_COM_MSG);
	// Puke(NULL_COM_MSG);
	scrollMsg(NULL_COM_MSG);
	return;
	}

// If the sensor has changed update it
if (lastSensorIdx != curCom->sensor_index)
	{
	lastSensorIdx = updateSensorName();
	}

// If the line has changed update it
if (strcmp(lineName, track->name))
	{
	strcpy(linTag->fld, track->name);
	DrawTaggedOb(&navDispFm->ob_lst[10]);
	strcpy(lineName, track->name);

	sprintf(rngTag->fld, "%.0lf", track->bearing * RAD_TO_DEG);
	DrawTaggedOb(&navDispFm->ob_lst[11]);

	sprintf(brgTag->fld, "%.0lf", track->length);
	DrawTaggedOb(&navDispFm->ob_lst[12]);
	}
}


/*----------------------------------------------------------------------------*/
// Save the Nav set
/*----------------------------------------------------------------------------*/
int write_nav(char * filename)
{
FILE *fp;
int k;

if ((fp = fopen(filename, "wb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  fclose(fp);
  return FALSE;
  }

k = fwrite( (void *) curNav, sizeof(t_nav), 1, fp);
if (k == -1)
	{
	fclose(fp);
	printf("\nfwrite failed");
	return FALSE;
	}

fclose(fp);
return TRUE;
}


/*----------------------------------------------------------------------------*/
// read the nav set
/*----------------------------------------------------------------------------*/
int read_nav(char * filename)
{
FILE *fp;
int k;
t_nav nav;

if ((fp = fopen(filename, "rb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  return FALSE;
  }

k = fread( (char *) &nav, sizeof(t_nav), 1, fp);
if (k == -1)
	{
	fclose(fp);
	printf("\nfread failed");
	return FALSE;
	}

// we're done with this so close it
fclose(fp);

// find the com structure that corresponds to this nav sensor type
curCom  = queryComByNav(nav.id);                  // Sensor Index
if (curCom == (t_com *) NULL)
	{
	scrollMsg(NULL_COM_MSG);
	printf("%s:%d %s ID:%d\n", __FILE__, __LINE__, NULL_COM_MSG, nav.id);
	// Puke(NULL_COM_MSG);
	return FALSE;
	}

// Now fetch the nav structure for this port
curNav = queryNavPtr(curCom->portnum);
if (curNav == (t_nav *) NULL)
	{
	NotifyUser(-1, -1, __FILE__, NULL_NAV_MSG);
	Puke(NULL_NAV_MSG);
	}

memcpy((void *) curNav, (void *) &nav, sizeof(t_nav));

if (_Debug)
	dumpNav();

return TRUE;
}



/*----------------------------------------------------------------------------*/
// read the nav set
/*----------------------------------------------------------------------------*/
void dumpNav(void)
{
int j = 0;
char buf[512];
time_t ltime = (long) curNav->tim;

j += sprintf(buf+j, "id = %d\n",     curNav->id);
j += sprintf(buf+j, "tim = %s", asctime(gmtime(&ltime)));
j += sprintf(buf+j, "lat = %.8lf\n", curNav->lat);
j += sprintf(buf+j, "lon = %.8lf\n", curNav->lon);
j += sprintf(buf+j, "spd = %.2lf\n", curNav->spd);
j += sprintf(buf+j, "crs = %.2lf\n", curNav->crs * RAD_TO_DEG);
j += sprintf(buf+j, "hdg = %.2lf\n", curNav->hdg * RAD_TO_DEG);

NotifyUser(-1, -1, NAV_FILE_NAME, buf);
}






/*----------------------------------------------------------------------------*/
// A simple key handler
/*----------------------------------------------------------------------------*/
int navKeyTask ( ob_desc * ob )
{
app_msg *key = ReReadMsgQueue ();
point pt;

pt.X = key->msgX;
pt.Y = key->msgY;

if ( isEscapeKey ( key ) ) 
	return ( USER_CANCEL );

// If it's not on make it on
ShowCursor();

switch( key->msgType )
	{
	case msDCLICK   :
		if ( (MouseInDrawingArea(key->msgX, key->msgY)) && 
         (OKtoDraw()) &&
		     (QuerySetup()->DisplayType != RANGE_REDUCTION_DISPLAY) )
			doRubberRect(ob);
		break;

	case msPRESS   :
		if ((OKtoDraw()) && (PtInWindow(&pt, videoWn)))  // click in video window?
			{
			SetActiveWn(videoWn, 0);
			// See if they clicked on lower right corner
			if (ptInVideoHotSpot(&pt))              // Resize it
				{
				ResizeWindow(1);
				}
			else                                    // Move it
				{
				MoveWindow();
				}

			updateVideoCoords();
			SetActiveWn(0, 0);
			}
		break;

	case kbKEYB:
		if ( check_for_print_screen ( key ) ) 
			return ( USER_IGNORE );
		break;

	case kbKEYP:
		switch (key->msgValue)
			{
			case kHOME:
				doSetCenter(ob);                // Set the center
				break;

			case kPGUP:
				incrementScale(.10);            // Increase the scale by 10 %
				break;

			case kPGDN:
				incrementScale(-.10);            // Decrease the scale by 10 %
				break;
			}

	case kbFUNC:
		switch(key->msgValue)
			{
			case F1_KEY:
				doLogBook(ob);
			break;

			case F2_KEY:
				doMarkPosition();
			break;

			case F7_KEY:
				process_inspection_port_key();
			break;
			}
		break;
	}

return ( USER_CONTINUE );
}





/*----------------------------------------------------------------------------*/
// update the static historical array of Ship position
/*----------------------------------------------------------------------------*/
void updateShipHistory(t_history * hist)
{
static double lastEntry;            // time of last historical entry
register int idx = hist->numPts;

// is current Nav OK?
if (validNav(curNav) != SUCCESS)
	return;

// is it time to add another entry?
if ((curNav->tim - lastEntry) >= (double) HISTORY_INTERVAL)
	{

	// is the array full ??
	if (idx == HISTORY_SIZE - 1)
		{
		// bump oldest element out, memmove preserves overlapping src and dest areas
		memmove((void *) &hist->fix[0], 
            (void *) &hist->fix[1], 
						sizeof (t_fix) * (HISTORY_SIZE - 1));
		}

	// keep adding it's not full yet
	hist->fix[idx].tim = lastEntry = curNav->tim;
	hist->fix[idx].pos.lat = curNav->lat;
	hist->fix[idx].pos.lon = curNav->lon;
	if (hist->numPts < HISTORY_SIZE - 1)
		hist->numPts++;
	}
}


/*----------------------------------------------------------------------------*/
// return a pointer to the ships history
/*----------------------------------------------------------------------------*/
t_history * queryShipHistory(void)
{
return (t_history *) &shipHistory;
}

/*----------------------------------------------------------------------------*/
// Handle Math errors instead of letting stderr do it
/*----------------------------------------------------------------------------*/
int matherr( struct exception *except )
{
// Handle DOMAIN errors for log or atan2. 
switch( except->type )
	{
	case DOMAIN:
		if( strcmp( except->name, "log" ) == 0 )
			{
			except->retval = log( -(except->arg1) );
			printf( "Special: using absolute value: %s: DOMAIN error\n",
                except->name );
			return 1;
			}

		if( strcmp( except->name, "atan2" ) == 0 )
			{
			except->retval = 0.0;
			printf( "Special: using absolute value: %s: DOMAIN error\n",
                except->name );
			return 1;
			}
		break;

	default:
		printf( "Normal:%s\n", except->name);
		return 0;    /* Else use the default actions */
		break;
	}
}




/*----------------------------------------------------------------------------*/
// Update the side panel with the appropriate form
/*----------------------------------------------------------------------------*/
void displayNav(void)
{
t_setup * config = QuerySetup();

if (config->SidePanelType == LINE_INFO)
	{
	updateLineInfo();
	}
else
	{
	updateSiteInfo();
	}
updateDepth();
updateDiskFree(0);
}



/*----------------------------------------------------------------------------*/
// Update the site info side panel
/*----------------------------------------------------------------------------*/
void updateSiteInfo(void)
{
static int lastSensorIdx = -1L;
static char siteName[30];
ftag * latTag;
ftag * lonTag;
ftag * timTag;
ftag * spdTag;
ftag * crsTag;
ftag * gyrTag;
ftag * navTag;
ftag * rngTag;
ftag * brgTag;
ftag * ttgTag;
ftag * namTag;
ftag * typTag;
ftag * infTag;

char buf[80];
int retv;
t_nav * gyroNav = (t_nav *) NULL;       // pointer to Gyro nav struct
t_rngBrg * site = (t_rngBrg *) NULL;    // pointer to site info

if ((DisplayingSomething) || navDispFm == (fm_desc *) NULL)
	{
	return;
	}
else
	{
	if ((retv = validNav(curNav)) != SUCCESS)
		return;
	}

timTag = (ftag *) navDispFm->ob_lst[0].tag;                // time of day GPS
latTag = (ftag *) navDispFm->ob_lst[1].tag;                // latitude
lonTag = (ftag *) navDispFm->ob_lst[2].tag;                // longitude
spdTag = (ftag *) navDispFm->ob_lst[3].tag;                // sped made good
crsTag = (ftag *) navDispFm->ob_lst[4].tag;                // course made good
gyrTag = (ftag *) navDispFm->ob_lst[5].tag;                // Gyro
navTag = (ftag *) navDispFm->ob_lst[6].tag;                // Nav Sensor Name
rngTag = (ftag *) navDispFm->ob_lst[7].tag;                // Range
brgTag = (ftag *) navDispFm->ob_lst[8].tag;                // Bearing
ttgTag = (ftag *) navDispFm->ob_lst[9].tag;                // Time to Go
namTag = (ftag *) navDispFm->ob_lst[10].tag;               // Site Name
typTag = (ftag *) navDispFm->ob_lst[11].tag;               // Site type    
infTag = (ftag *) navDispFm->ob_lst[11].tag;               // Site Info??
	
PositionString(lonTag->fld, latTag->fld, buf, 
               curNav->lon, curNav->lat, 0.0, 0);

strcpy(timTag->fld, getTimeStr(0.0));
sprintf(spdTag->fld, "%5.1lf", curNav->spd);
sprintf(crsTag->fld, "%5.1lf", curNav->crs * RAD_TO_DEG);

// get the active gyro if there is one
gyroNav = queryNavByType(QuerySetup()->Gyro);
if (gyroNav == (t_nav *) NULL)
	{
	// must not be a gyro connected so use the crs made good
	sprintf(gyrTag->fld, "%5.1lf", curNav->crs * RAD_TO_DEG);
	// scrollMsg(NULL_COM_MSG);
	}
else
	{
	curNav->hdg = gyroNav->hdg;
	sprintf(gyrTag->fld, "%5.1lf", (curNav->hdg = gyroNav->hdg) * RAD_TO_DEG);
	}

if ((site = querySiteInfo()) == (t_rngBrg *) NULL)
	{
	scrollMsg("Null Site\n");
	return;
	}
else
	{
	sprintf(rngTag->fld, "%6.0lf", site->range);
	sprintf(brgTag->fld, "%5.1lf", site->bearing * RAD_TO_DEG);
	}

DrawTaggedOb(&navDispFm->ob_lst[0]);
DrawTaggedOb(&navDispFm->ob_lst[1]);
DrawTaggedOb(&navDispFm->ob_lst[2]);
DrawTaggedOb(&navDispFm->ob_lst[3]);
DrawTaggedOb(&navDispFm->ob_lst[4]);
DrawTaggedOb(&navDispFm->ob_lst[5]);

DrawTaggedOb(&navDispFm->ob_lst[7]);
DrawTaggedOb(&navDispFm->ob_lst[8]);
DrawTaggedOb(&navDispFm->ob_lst[9]);

// Use the sensor index to get the associated port
if (curCom == (t_com *) NULL)
	{
	scrollMsg(NULL_COM_MSG);
	return;
	}

// If the sensor has changed update it
if (lastSensorIdx != curCom->sensor_index)
	{
	lastSensorIdx = updateSensorName();
	}

// If the site has changed update it
if (strcmp(siteName, site->site.name))
	{
	strcpy(namTag->fld, site->site.name);
	DrawTaggedOb(&navDispFm->ob_lst[10]);
	strcpy(siteName, site->site.name);

	strcpy(typTag->fld, querySampleTypeName(site->site.type));
	DrawTaggedOb(&navDispFm->ob_lst[11]);
	}
}

/*----------------------------------------------------------------------------*/
// update the sensor name on the side panel display
/*----------------------------------------------------------------------------*/
int updateSensorName(void)
{
ftag * navTag = (ftag *) navDispFm->ob_lst[6].tag;               // Site Name

// Use the sensor index to get the associated port
if (curCom == (t_com *) NULL)
	{
	scrollMsg(NULL_COM_MSG);
	return 0;
	}

strcpy(navTag->fld, curCom->sensor_name);
DrawTaggedOb(&navDispFm->ob_lst[6]);

return curCom->sensor_index;
}


/*----------------------------------------------------------------------------*/
// update the Depth on side panel display
/*----------------------------------------------------------------------------*/
void updateDepth(void)
{
t_setup * config = QuerySetup();
ftag * tag;

tag = (ftag *) navDispFm->ob_lst[14].tag;
sprintf(tag->fld, "%.0lf%c", curNav->alt, (depthAlarmFlag) ? '*' : ' ');
DrawTaggedOb(&navDispFm->ob_lst[14]);
}



/*----------------------------------------------------------------------------*/
// update the amount of disk space available on the log drive
/*----------------------------------------------------------------------------*/
void updateDiskFree(int forceIt)
{
ftag * tag;
t_setup * config = QuerySetup();
time_t now = time(NULL);
static time_t lastTime = 0L;
unsigned int drive;

if ((forceIt) || (now - lastTime > (time_t) 20))
	{
	if ((DisplayingSomething) || navDispFm == (fm_desc *) NULL)
		{
		return;
		}

	tag = (ftag *) navDispFm->ob_lst[13].tag;
	if ((config->BinLoggingEnabled)       ||
      (config->AsciiLogEnabled))
		{
		drive = getDriveNumber(config->binFilename);
		sprintf(tag->fld, "%c: %lu", 
            drive + 'A' - 1, 
            diskFree(getDriveNumber(config->binFilename)));
		}
	else
		{
		strcpy(tag->fld, "Logging OFF!");
		}
	DrawTaggedOb(&navDispFm->ob_lst[13]);
	lastTime = now;
	}
}
