/*----------------------------------------------------------------------------*/
// File: offsets.c   USGS Navigation
// Author: John Gann
// Start Date: 10 April 1992  Sensor Offset functions 
/*----------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <time.h>
#include <math.h>
#include <limits.h>

#include <GRconst.h>
#include <GRports.h>
#include <GRextrn.h>
#include <menuet.h>

#include "navdefs.h"
#include "navtypes.h"
#include "nav_util.h"
#include "graph.h"
#include "mnutils.h"
#include "offsets.h"

// Private data
static char * CLEAR_MSG = "                                                   ";
static char * NULL_SETUP_MSG  = "Null Configuration";
static char * NULL_COM_MSG    = "Null Com Port";
static char * NULL_TRACK_MSG    = "Null Track Structure";
static t_pos VrpPos;                           // Vehicle Reference position
static t_offset Offsets[MAX_SENSORS];     // Offsets of each sensor

// External data
extern t_nav * curNav;      // pointer to the current navigation 
extern char * SensorNames[];

/*----------------------------------------------------------------------------*/
// read the offsets setup
/*----------------------------------------------------------------------------*/
int read_offsets(char * filename)
{
FILE *fp;
int i, k, j=0;

if ((fp = fopen(filename, "rb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  fclose(fp);
  return FALSE;
  }

for (i= 0; i < MAX_SENSORS; i++)
	{
	k = fread( (void *) &Offsets[i], sizeof(t_offset), 1, fp);
	if (k == -1)
		{
		fclose(fp);
		printf("\nfread failed");
		return FALSE;
		}
	}

fclose(fp);
return TRUE;
}





/*----------------------------------------------------------------------------*/
// Save the Offsets file 
/*----------------------------------------------------------------------------*/
int write_offsets(char * filename)
{
FILE *fp;
int i, k;

if ((fp = fopen(filename, "wb")) == NULL)
  {
  printf("\nCan't open file %s", filename);
  fclose(fp);
  return FALSE;
  }

for (i= 0; i < MAX_SENSORS; i++)
  {
  k = fwrite( (char *) &Offsets[i], sizeof(t_offset), 1, fp);
  if (k == -1)
    {
    fclose(fp);
    printf("\nfwrite failed");
    return FALSE;
    }
  }

fclose(fp);
return TRUE;
}


/*----------------------------------------------------------------------------*/
// do the Offsets form
/*----------------------------------------------------------------------------*/
int doOffsetsFm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "offset.frf";
ftag * xTag;
ftag * yTag;
ftag * zTag;
ltag * sensorTag;
static int lastTag = 0;

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

// set up form
xTag      = (ftag *) fm->ob_lst[2].tag;
yTag      = (ftag *) fm->ob_lst[3].tag;
zTag      = (ftag *) fm->ob_lst[4].tag;
sensorTag = (ltag *) fm->ob_lst[5].tag;
sensorTag->curitem = lastTag;

sprintf(xTag->fld, "%.0lf",  Offsets[sensorTag->curitem].x);
sprintf(yTag->fld, "%.0lf",  Offsets[sensorTag->curitem].y);
sprintf(zTag->fld, "%.0lf",  Offsets[sensorTag->curitem].z);

OpenForm ( fm, ob->area.Xmin-40, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the fields
			lastTag = sensorTag->curitem;
			Offsets[lastTag].x = atof(xTag->fld);
			Offsets[lastTag].y = atof(yTag->fld);
			Offsets[lastTag].z = atof(zTag->fld);

			// compute the range and bearing to the Stern Center
			updateOffsetR_B((t_offset *) &Offsets[lastTag]);
			break;

		case CANCEL:
			break;

		default : 
			updateOffsetFm(fm);
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}


/*----------------------------------------------------------------------------*/
// update the offset form
/*----------------------------------------------------------------------------*/
void updateOffsetFm(fm_desc * fm)
{
ftag * xTag      = (ftag *) fm->ob_lst[2].tag;
ftag * yTag      = (ftag *) fm->ob_lst[3].tag;
ftag * zTag      = (ftag *) fm->ob_lst[4].tag;
ltag * sensorTag = (ltag *) fm->ob_lst[5].tag;

// update the x, y and z for the current sensor
sprintf(xTag->fld, "%.0lf",  Offsets[sensorTag->curitem].x);
sprintf(yTag->fld, "%.0lf",  Offsets[sensorTag->curitem].y);
sprintf(zTag->fld, "%.0lf",  Offsets[sensorTag->curitem].z);

// redraw the fields
DrawTaggedOb(&fm->ob_lst[2]);
DrawTaggedOb(&fm->ob_lst[3]);
DrawTaggedOb(&fm->ob_lst[4]);
}


/*----------------------------------------------------------------------------*/
// do the Vehicle Reference Position form
/*----------------------------------------------------------------------------*/
int doVrpFm ( ob_desc * ob )
{
int retv;
fm_desc * fm;
char * frf = "vrp.frf";
ftag * xTag;
ftag * yTag;
ftag * zTag;
t_setup * config = QuerySetup();

if (config == (t_setup *) NULL)
	{
	scrollMsg(NULL_SETUP_MSG);
  return BAD_SETUP;
	}

// Load the FRF file 
if ( ( fm = loadFRF ( frf ) ) == NULL ) 
	{
  return FALSE;
	}

// set up form
xTag      = (ftag *) fm->ob_lst[2].tag;
yTag      = (ftag *) fm->ob_lst[3].tag;
zTag      = (ftag *) fm->ob_lst[4].tag;

sprintf(xTag->fld, "%.0lf",  config->Vrp.x);
sprintf(yTag->fld, "%.0lf",  config->Vrp.y);
sprintf(zTag->fld, "%.0lf",  config->Vrp.z);

OpenForm ( fm, ob->area.Xmin-40, ob->area.Ymin ); 

do 
	{
	retv = ProcessObList ( fm->ob_lst, fm->ob_cnt, 0 );
	switch ( retv ) 
		{
		case OKAY:
			// get the fields
			config->Vrp.x = atof(xTag->fld);
			config->Vrp.y = atof(yTag->fld);
			config->Vrp.z = atof(zTag->fld);
			updateOffsetR_B((t_offset *) &config->Vrp);
			break;

		case CANCEL:
			break;
		}
	} while (retv != OKAY && retv != CANCEL - 1);

CloseForm   ( fm );
DestroyFRFForm ( fm );

return retv;
}

/*----------------------------------------------------------------------------*/
// update the offsets range and bearing
/*----------------------------------------------------------------------------*/
void updateOffsetR_B(t_offset *p)
{
p->dist = hypot(p->x, p->y);

if (p->dist != 0.0)
	p->bearing = atan2(p->x, p->y);
else
	p->bearing = 0.0;

return;
}


/*----------------------------------------------------------------------------*/
// compute the geographical coordinates of the vehicle reference position
/*----------------------------------------------------------------------------*/
int computeVRP(t_nav * p)
{
double bearing;
int retv;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid 
t_offset * off = &Offsets[p->id];

// sanity checks
if (off == (t_offset *) NULL)
	return  BAD_OFFSET;

// compute the bearing from the projection. 
// i.e. normal rotation offset plus the current gyro
bearing = fmod(p->hdg + off->bearing, TWOPI);

// chack the nav pointer
if ((retv = validNav(p)) != SUCCESS)
	{
	return retv;
	}

// compute position at VRP
retv = proj(p->lat, p->lon, 
            off->dist, bearing, 
            &VrpPos.lat, &VrpPos.lon,
            ellipse->a, ellipse->b);

return retv;
}
			

/*----------------------------------------------------------------------------*/
// compute the geographical coordinates of the passed offset
/*----------------------------------------------------------------------------*/
int computeOffset(int sensorIdx, double * lat, double * lon)
{
double bearing;
// char buf[256];
int retv;
t_ellips * ellipse = query_ellipsoid();   // ellipsoid 
t_offset * off = &Offsets[sensorIdx];

// sanity checks
if (off == (t_offset *) NULL)
	return  BAD_OFFSET;

// compute the bearing from the projection. 
// i.e. normal rotation offset plus the current gyro
bearing = fmod(curNav->hdg + off->bearing, TWOPI);

// compute position at VRP
retv = proj(VrpPos.lat, VrpPos.lon, 
            off->dist, bearing, 
            lat, lon,
            ellipse->a, ellipse->b);

/*******
sprintf(buf, "%s:%d computeOffset() retv %d\nvrpLat %.9lf vrpLon %.9lf\nrange %.1lf bearing %.1lf off->bearing %.1lf gyro %.1lf\nnew lat %.9lf new lon %.9lf\n", __FILE__, 
__LINE__, 
retv, 
VrpPos.lat, 
VrpPos.lon, 
off->dist, 
bearing      * RAD_TO_DEG, 
off->bearing * RAD_TO_DEG, 
curNav->hdg  * RAD_TO_DEG, 
*lat, 
*lon);

logMsg(buf);
*******/

return retv;
}


/*----------------------------------------------------------------------------*/
// return a pointer to the VRP
/*----------------------------------------------------------------------------*/
t_pos * queryVRP(void)
{
return (t_pos *) &VrpPos;
}

/*----------------------------------------------------------------------------*/
// return a pointer to the requested Offset struct
/*----------------------------------------------------------------------------*/
t_offset * queryOffset(int which)
{
if ( (which < sizeof(Offsets) / sizeof(Offsets[0])) && 
     (which >= 0) )
	return (t_offset *) &Offsets[which];
else
	return (t_offset *) NULL;
}
