/*----------------------------------------------------------------------------*/
// Name: stans.c
// Blame: John Gann
// Start Date: 24 Sep 92
// Functions to communicate with the Trimble TANS and Super TANS family
/*----------------------------------------------------------------------------*/
#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <conio.h>
#include <dos.h>

#include <GRconst.h>
#include <GRports.h>
#include <GRextrn.h>
#include <menuet.h>

#include "tansdefs.h"
#include "tanstype.h"
#include "nav_utils.h"
#include "serial.h"
#include "ec_comm.h"
#include "graph.h"

#include "stans.h"

// Private Data
static t_tans theTans;            
static char _Debug              = FALSE;
static char _Verbose            = FALSE;
static char errBuf[256];
static unsigned int   RX_count = 0;
static unsigned char  HaveDLE = FALSE;
static unsigned int AwaitingQuery = 0;
static void (*queryCallBackFunc)(unsigned int) = NULL;

#define getChar get_char
#define putChar put_char

/*----------------------------------------------------------------------------*/
// initialize the TANS structure and library
/*----------------------------------------------------------------------------*/
int initTans(int port)
{
// Initialize the TANS structure
theTans.cfg.com.port           = port;
theTans.cfg.com.baud           = 9600;
theTans.cfg.com.parity         = 'O';
theTans.cfg.com.data           = 8;
theTans.cfg.com.stop           = 1;

theTans.cfg.lastErr        = 0;
theTans.cfg.com.timeOfLastPkt  = (time_t) 0;

// Set control lines high
setControlLines(port, 1, 1);

// Set Tans in a 1 sec sync mode
setTansIoOptions();

// Initialize the TANS structure
theTans.cfg.initialized    = TRUE;

return SUCCESS;
}



/*----------------------------------------------------------------------------*/
// display the most recent TANS error
/*----------------------------------------------------------------------------*/
char *  queryTansError(void)
{
static char * tansErrors[] = 
	{
	"No Error",
	"Bad Position"
	};

if ((theTans.cfg.lastErr >= 0) && 
    (theTans.cfg.lastErr < sizeof(tansErrors) / sizeof(tansErrors[0])))
	return tansErrors[theTans.cfg.lastErr];

return "Unknown Error";
}



/*----------------------------------------------------------------------------*/
// return a pointer to the static tans struct
/*----------------------------------------------------------------------------*/
t_tans * queryTansStruct(void)
{
return (t_tans *) &theTans;
}




/*----------------------------------------------------------------------------*/
// update the static tans structure and process packets
/*----------------------------------------------------------------------------*/
int updateTans(void)
{
int retv;

if (theTans.cfg.initialized)
	retv = processTansPackets(theTans.cfg.com.port, 5);
else
	retv = TANS_NOT_INITIALIZED;

return (retv == NEW_INFO) ? SUCCESS : -1;
}





/*----------------------------------------------------------------------------*/
// Stream positions as fast as they are computed
/*----------------------------------------------------------------------------*/
int setTansIoOptions(void)
{
putChar(theTans.cfg.com.port, (unsigned char) DLE);
putChar(theTans.cfg.com.port, (unsigned char) SET_IO_OPTIONS);
putChar(theTans.cfg.com.port, (unsigned char) 0x12);  	// Dbl Precision & HAE
putChar(theTans.cfg.com.port, (unsigned char) 0x02);    // velocity ENU On
putChar(theTans.cfg.com.port, (unsigned char) 0x02);    // Fix next int second
putChar(theTans.cfg.com.port, (unsigned char) 0x02);    // MSMT, req only & smoothed
putChar(theTans.cfg.com.port, (unsigned char) DLE);
putChar(theTans.cfg.com.port, (unsigned char) ETX);
return SUCCESS;
};



/*----------------------------------------------------------------------------*/
// Build a packet returning false until packet is complete                    
/*----------------------------------------------------------------------------*/
int build_packet (int com_port, unsigned int *idcode, unsigned char pktbuff[])
{
static unsigned char this_char;

if (RX_count > 250)
	{
	printf("\nYo!, Rxcnt %d\n\n", RX_count);
	RX_count = 0;
	return TRUE;
	}

if (!bytesWaiting(com_port))
	return FALSE;

if ((this_char = getChar( com_port)) == -1) 
	{
	return -1;
	}

if (!HaveDLE)
	{
	if (this_char == DLE)
		{
		HaveDLE = TRUE;
		return FALSE;
		}

	if (RX_count < 250)
		{
		pktbuff[RX_count++] = this_char;
		}
	return FALSE;
	}

if (this_char == DLE)
	{
	if (RX_count < 250)
		{
		pktbuff[RX_count++] = DLE;
		}
	HaveDLE = FALSE;
	return FALSE;
	}

if (this_char != ETX)
	{
	RX_count = 0;
	*idcode = this_char;
	HaveDLE = FALSE;
	return FALSE;
	}

return RX_count;
}




/*----------------------------------------------------------------------------*/
/* Reset packet builder                                                       */
/*----------------------------------------------------------------------------*/
void reset_packet_builder( void )
{
HaveDLE = FALSE;
RX_count = 0;
}




/*----------------------------------------------------------------------------*/
// Process TANS Packets, for a period of time                                 
/*----------------------------------------------------------------------------*/
int processTansPackets(int com_port, int seconds_to_process)
{
char buf[256];
static unsigned int pkt_id;
static unsigned char pktbuff[1024];
int pkt_size;
char display_message;
int bytes_waiting = bytesWaiting(com_port);
clock_t idle_time = 0L;
clock_t start_time = clock();                  // time whole thing started
clock_t last_time = start_time;                  // time of last bytes_waiting
int retval = PACKET_IN_PROGRESS;

do
	{
	bytes_waiting = bytesWaiting(com_port);
	if (bytes_waiting == 0)
		{
		idle_time = clock() - last_time;
		}
	else
		{
		idle_time = 0L;
		last_time = clock();
		}

	if (pkt_size = build_packet (com_port, &pkt_id, pktbuff))
		{
		theTans.cfg.com.timeOfLastPkt = time(NULL);
		if (_Verbose)
			{
			sprintf(buf, "PKT: %X\nSize: %X(hex)\n", pkt_id, pkt_size);
			scrollMsg(buf);
			}

		display_message = TRUE;
		memset(buf, 0x00, sizeof(buf));

		switch(pkt_size)
			{
			case 0:
			case -1:
				sprintf(buf, "Timeout\n");
				scrollMsg(buf);
				break;
		
			default:	
				switch (pkt_id)
					{
					case POS_LLA_DBL_PKT:
						strcpy(buf, "Position Fix LAT LON ALT\n");
						if (pkt_size == 36)
							{
         			processPositionPacket(pktbuff);
							retval = NEW_INFO;
							}
						break;
		
					case SAT_DOP_MODE_PKT:
	    			if (pkt_size == 21)  
							{
							processSatDopMode(pktbuff, pkt_size);
							}
						break;
		
					case MACHINE_CODE_PKT:
						processMachineCode(pktbuff);
						// scrollMsg(qryTansMachineCode());
   					break;

					case GPS_SYS_TIME_PKT:
						strcpy(buf, "GPS Time\n");
						processSysTime(pktbuff);
						// scrollMsg(qryTansTime());
						break;
		
					case VEL_FIX_ENU_PKT:
						strcpy(buf, "Velocity\n");
						break;
		
					case HEALTH_PKT:                         /* 0x46 */
						display_message = FALSE;
						processHealth(pktbuff);
						// scrollMsg(qryGpsHealth());
						// scrollMsg(qryTansStatus());
						break;

					case SOFT_VER_PKT:
						display_message = FALSE;
						processFirmwarePkt(pktbuff);
						break;

					case OP_PRAMS_PKT:
						if (pkt_size == 17)
							{
							processOpPramsPkt(pktbuff);
							}
   					break;

					default:
						sprintf(buf, "PKT: %X\nSize: %X(hex)\n\n", pkt_id, pkt_size);
						break;
					}   /* end of switch on pkt_id */
				break;  
			}  /* end of switch on pkt_size */

		// If a query was pending now is the time to notify the caller
		if (AwaitingQuery == pkt_id)
					replyToQuery(pkt_id);

		if ((display_message) && (buf[0] != '\0'))
			scrollMsg(buf);

		reset_packet_builder();
		}    /* end of if pkt_size */

	} while (idle_time <= 2L);   // 2 ticks (.125s idle time allowed)

return retval;
}
	



/*----------------------------------------------------------------------------*/
/* Process_pos_lla_dbl returns nonzero if valid fix                           */
/*----------------------------------------------------------------------------*/
int processPositionPacket(unsigned char * pktbuff)
{
static int first_fix = TRUE;
static float lastTime = 0.0F;
double lat;
double lon;
double alt;
float  tim;

/* Always ignore the first fix */
if (first_fix)
	{
	first_fix = FALSE;
	return FALSE;
	}

if (theTans.gps.health == HIGH_PDOP)
	return BAD_GPS_HEALTH;

lat        = swapDouble(&pktbuff[0]);
lon        = swapDouble(&pktbuff[8]);
alt        = swapDouble(&pktbuff[16]);
tim        = swapFloat (&pktbuff[32]);

if ( tim == lastTime )
	return OLD_FIX_DATA;

if ( tim < 0.0F      || tim > (float) SECONDS_PER_WEEK
	|| lat < -HALFPI   || lat > HALFPI
	|| lon < -TWOPI    || lon > TWOPI
	|| alt < -AE_WGS84 || alt > AE_WGS84)
	{
	scrollMsg("BAD Fix data \n");
	return BAD_FIX_DATA;
	}

lastTime = tim;
theTans.gps.pos.lla.lat = (float) lat;
theTans.gps.pos.lla.lon = (float) lon;
theTans.gps.pos.lla.alt = (float) alt;
theTans.gps.timeOfFix   = (double) tim;

return SUCCESS;
}


/*----------------------------------------------------------------------------*/
// Process TANS packet 0x44 SAT_DOP_MODE                                      
/*----------------------------------------------------------------------------*/
int processSatDopMode(char pktbuff[], int pktsize)
{
int i, j = 0;
char buf[80];
int temp;
char fix_mode = pktbuff[0];
int satCnt = 0;
int prnsUsed[5];

// clear out crap
memset((void *) prnsUsed, 0x00, sizeof(prnsUsed));

for(i=1; i<5; i++)
	{
	temp = (int) pktbuff[i];
	if (temp >= 0 && temp <= 32)
		{
		if (temp != 0)
			satCnt++;
		prnsUsed[i-1] = pktbuff[i];
		j += sprintf(buf+j, "SV%02d ", pktbuff[i]);
		}
	else
		{
		sprintf(buf, "Bad Sat Dop Fixmode %d SV's %02d, %02d, %02d, %02d\n", fix_mode, pktbuff[1], pktbuff[2], pktbuff[3], pktbuff[4]);
		scrollMsg(buf);
		return BAD_SAT_DOP_MODE;
		}
	}

memcpy((void *) theTans.gps.prnsUsed, 
       (void *) prnsUsed,
       sizeof(theTans.gps.prnsUsed));

theTans.gps.numSatsUsed = satCnt;
theTans.gps.pdop = fabs((double) swapFloat( &pktbuff[5] ));
theTans.gps.hdop = fabs((double) swapFloat( &pktbuff[9] ));
theTans.gps.vdop = fabs((double) swapFloat( &pktbuff[13] ));

strcat(buf, "\n");
scrollMsg(buf);

return SUCCESS;
}






/*----------------------------------------------------------------------------*/
// Swap a float from Motorola to Intel format                                 
/*----------------------------------------------------------------------------*/
float swapFloat(char * p)
{
t_uval val;

val.ch[3] = *p++;
val.ch[2] = *p++;
val.ch[1] = *p++;
val.ch[0] = *p;

return(val.f);
} 



/*----------------------------------------------------------------------------*/
// Swap an int from Motorola to Intel format                                  
/*----------------------------------------------------------------------------*/
int	swapInt(char *p)
{
t_uval val;

val.ch[1] = *p++;
val.ch[0] = *p;

return(val.i);
}



/*----------------------------------------------------------------------------*/
// Swap a double from Motorola to Intel format                                
/*----------------------------------------------------------------------------*/
double	swapDouble (char *p)
{
t_uval uval;

uval.ch[7] = *p++;
uval.ch[6] = *p++;
uval.ch[5] = *p++;
uval.ch[4] = *p++;
uval.ch[3] = *p++;
uval.ch[2] = *p++;
uval.ch[1] = *p++;
uval.ch[0] = *p;

return(uval.d);
} 



/*----------------------------------------------------------------------------*/
// shut down the TANS comm link but leave DTR line high to keep receiver hot
/*----------------------------------------------------------------------------*/
void termTans(char * msg)
{
}


/*----------------------------------------------------------------------------*/
// set the static debug and verbose flags
/*----------------------------------------------------------------------------*/
void setTansDebugVerbose(int debug, int verbose)
{
_Debug   = debug;
_Verbose = verbose;
}


/*----------------------------------------------------------------------------*/
// Return a pointer to fix mode string
/*----------------------------------------------------------------------------*/
char * queryFixMode(void)
{
static char * modeLabels[] = 
	{
	"Auto 2D",
	"Auto 3D",
	"Manual 2D",
	"Manual 3D"
	};

if ((theTans.gps.fixMode >= 0) && 
    (theTans.gps.fixMode < sizeof(modeLabels) / sizeof(modeLabels[0])))
	return modeLabels[theTans.gps.fixMode];

return "Unknown Fix Mode";
}




/*----------------------------------------------------------------------------*/
/* Process TANS Packet 0x4B Machine Code ID and additional Status             */
/* Return  1  if valid and different from last                                */
/*----------------------------------------------------------------------------*/
int processMachineCode ( unsigned char pktbuff[] )
{
theTans.gps.machineCodeId = pktbuff[0];
return SUCCESS;
}




/*----------------------------------------------------------------------------*/
// return pointer to machine code name
/*----------------------------------------------------------------------------*/
char * qryTansMachineCode(void)
{
static char	*machineCodeLabels[] = 
      { 
      "10X and 400", 
      "10X", 
      "400", 
      "Datum Timing Module",
			"2-Chan TANS", 
      "6-Chan TANS", 
      "2-Chan TRIMPACK",
			"TANS Timing Module", 
      "ND TANS", 
      "3-Chan TANS",
			"3 or 6 Chan TRIMPACK", 
      "Navigator GPS Sensor",
      "Aircraft GPS Sensor", 
      "Transpack", 
      "Pathfinder Basic",
      "Transpack 2", 
      "??", 
      "3-Chan Smart Antenna", 
      "Acutis", 
      "DAR GPS Rcvr", 
      "6-Chan TANS" };

if ((theTans.gps.machineCodeId >= 0) && 
    (theTans.gps.machineCodeId < sizeof(machineCodeLabels) / 
                             sizeof(machineCodeLabels[0])))
	return machineCodeLabels[theTans.gps.fixMode];

return "Unknown Unit";
}	




/*----------------------------------------------------------------------------*/
// Process the GPS health packet
/*----------------------------------------------------------------------------*/
void processHealth(unsigned char pktbuff[])
{
theTans.gps.health = (unsigned char) pktbuff[0];
theTans.gps.status = (unsigned char) pktbuff[1];
}




/*----------------------------------------------------------------------------*/
// Return String of GPS health
/*----------------------------------------------------------------------------*/
char *  qryGpsHealth(void)
{
static char	* tansStatusLabels[] = 
	{ 
	"Doing position fixes",
	"Don't have GPS time yet",
	"Waiting for almanac collection",
	"PDOP too high",
	"No satellites available",
	"Only 1 satellite available",
	"Only 2 satellites available",
	"Only 3 satellites available",
	"No satellites usable",
	"Only 1 usable satellite ",
	"Only 2 usable satellites ",
	"Only 3 usable satellites",
	"Chosen satellite unusable"  
	};

if (theTans.gps.health > sizeof(tansStatusLabels) / sizeof(tansStatusLabels[0]))
	return "Unknown";

return tansStatusLabels[theTans.gps.health];
}



/*----------------------------------------------------------------------------*/
// Return String of Tans Status
/*----------------------------------------------------------------------------*/
char * qryTansStatus(void)
{
int i;
static char	* tansErrLabels[] = 
	{ 
	"Battery back-up failed",
	"Signal processor error",
	"Alignment of channel 1",
	"Alignment of channel 2",
	"Possible Antenna feed line fault",
	"Excessive Bias Dot"  
	};
char buffer[80];
int err = 0;

if (theTans.gps.health > sizeof(tansErrLabels) / sizeof(tansErrLabels[0]))
	return "Unknown";

for (i=0; i<6; i++)
	{
	if (theTans.gps.status & (1 << i))
		{
		err = TRUE;
    sprintf(buffer, "%xh %s\n", theTans.gps.status, tansErrLabels[i]);
		}
	}

if (err)
	{
	return (char *) buffer;
	}

return "Status: Ok";
}





/*----------------------------------------------------------------------------*/
// Process TANS PKT 0x41 System Time                                          
/*----------------------------------------------------------------------------*/
int processSysTime( unsigned char pktbuff[])
{
int gpsWeek;
float tow;

tow     = swapFloat ( &pktbuff[0] );
gpsWeek = swapInt ( &pktbuff[4] );

if (tow < 0.0F)
	{
	return BAD_TIME_PKT;
	}
else
	{	
	theTans.gps.gpsTow  = (double) tow;
	theTans.gps.gpsWeek = gpsWeek;
	}

return SUCCESS;
}



/*----------------------------------------------------------------------------*/
// Return String of GPS Time
/*----------------------------------------------------------------------------*/
char * qryTansTime(void)
{
return getTimeStr(GPStoDOS(theTans.gps.gpsTow, theTans.gps.gpsWeek));
}

/*----------------------------------------------------------------------------*/
// Return String of Last fix time
/*----------------------------------------------------------------------------*/
char * qryLastFixTime(void)
{
return getTimeStr(GPStoDOS(theTans.gps.timeOfFix, theTans.gps.gpsWeek));
}


/*----------------------------------------------------------------------------*/
// Set the AwaitingQuery packet ID
/*----------------------------------------------------------------------------*/
void setActiveQuery(int pktId)
{
AwaitingQuery = pktId;
switch(pktId)
	{
	case OP_PRAMS_PKT:
		setReqGpsOperPrams (0, -1.0F, -1.0F, -1.0F, -1.0F );
		break;

	case SOFT_VER_PKT:
		requestSimplePkt(REQ_FIRMWARE);
		break;
	}
}



/*----------------------------------------------------------------------------*/
// Clear the AwaitingQuery packet ID
/*----------------------------------------------------------------------------*/
void clearActiveQuery(void)
{
AwaitingQuery = 0;
}



/*----------------------------------------------------------------------------*/
// Return the packet ID that we are looking for
/*----------------------------------------------------------------------------*/
int queryActiveQuery(void)
{
return AwaitingQuery;
}



/*----------------------------------------------------------------------------*/
// Send/Request GPS operating parameters to TANS
/*----------------------------------------------------------------------------*/
int setReqGpsOperPrams(int dynamics, 
                       double el_mask, 
                       double snr_mask, 
                       double PDOP_mask, 
                       double PDOP_switch )
{
putChar(theTans.cfg.com.port, DLE);
putChar(theTans.cfg.com.port, SET_OPERATING_PARAMS);

putChar(theTans.cfg.com.port, (unsigned char) dynamics);

sendFloat((float) el_mask);
sendFloat((float) snr_mask);
sendFloat((float) PDOP_mask);
sendFloat((float) PDOP_switch);

putChar(theTans.cfg.com.port, DLE);
putChar(theTans.cfg.com.port, ETX);

return TRUE;
}




/*----------------------------------------------------------------------------*/
// Send the TANS an Intel Float by Converting it to Motorloa Format           
/*----------------------------------------------------------------------------*/
void sendFloat(float a)
{
int i;
t_uval uval;

uval.f = a;
for (i=3; i>=0; i--)
	{
  putChar(theTans.cfg.com.port, uval.ch[i]);
	if (uval.ch[i] == DLE)
	  putChar(theTans.cfg.com.port, uval.ch[i]);
	}	
}




/*----------------------------------------------------------------------------*/
// process the GPS Operational parameters packet
/*----------------------------------------------------------------------------*/
void processOpPramsPkt(unsigned char pktbuff[])
{
theTans.cfg.opPrams.dynamics  = (unsigned char) pktbuff[0];
theTans.cfg.opPrams.elvMask   = swapFloat (&pktbuff[1]);           // radians
theTans.cfg.opPrams.snrMask   = swapFloat (&pktbuff[5]);
theTans.cfg.opPrams.dopMask   = swapFloat (&pktbuff[9]);
theTans.cfg.opPrams.dopSwitch = swapFloat (&pktbuff[9]);
}


/*----------------------------------------------------------------------------*/
// replyToQuery notifies the caller through the caller specified call back
// function.  If no call back function has been specified, then the caller will
// not be notified that his query has been answered.
// In any event theTans structure will be updated and the AwaitingQuery
/*----------------------------------------------------------------------------*/
void replyToQuery(unsigned int pkt_id)
{
// Clear the query packet
clearActiveQuery();

// If the caller has specified a callback function, call him with the packet ID
if (queryCallBackFunc != NULL)
	{
	(*queryCallBackFunc)(pkt_id);
	}
}



/*----------------------------------------------------------------------------*/
// Set a function pointer to the query call back
// If you don't want to be called back, set the queryCallBack func to NULL
/*----------------------------------------------------------------------------*/
void setQueryCallBackFunc(void (*callBackFunc)(unsigned int))
{
queryCallBackFunc = callBackFunc;
}


/*----------------------------------------------------------------------------*/
// Clear the queryCallBackFunc pointer to NULL
/*----------------------------------------------------------------------------*/
void clearQueryCallBackFunc()
{
queryCallBackFunc = NULL;
}

/*----------------------------------------------------------------------------*/
// Request Any Simple TANS Packet that requires no data
/*----------------------------------------------------------------------------*/
void requestSimplePkt(unsigned char thePacket)
{
putChar(theTans.cfg.com.port, DLE);
putChar(theTans.cfg.com.port, thePacket);
putChar(theTans.cfg.com.port, DLE);
putChar(theTans.cfg.com.port, ETX);
}


/*----------------------------------------------------------------------------*/
// Process a TANS Firmware packet
/*----------------------------------------------------------------------------*/
void processFirmwarePkt(unsigned char * thePacket)
{
memcpy((void *) &theTans.cfg.firmware,
       (void *) thePacket,
       sizeof(theTans.cfg.firmware));
}


/*----------------------------------------------------------------------------*/
// Process a TANS Current Route Packet 0xC3 CURRENT_ROUTE_PKT
/*----------------------------------------------------------------------------*/
void processCurrentRoute(unsigned char * thePacket)
{
theTans.nav.curRoute       = (unsigned char) thePacket[0];
theTans.nav.curLeg         = (unsigned char) thePacket[1];
theTans.nav.nextWaypoint   = swapInt(&thePacket[2]);
}



/*----------------------------------------------------------------------------*/
// Process a TANS Current Route Packet 0xC5 CURRENT_DATUM_PKT
/*----------------------------------------------------------------------------*/
void processCurrentDatum(unsigned char * thePacket)
{
theTans.cfg.datum.id    = (unsigned char) thePacket[0];
theTans.cfg.datum.a     = swapFloat(&thePacket[1]);
theTans.cfg.datum.e     = swapFloat(&thePacket[5]);
theTans.cfg.datum.dx    = swapFloat(&thePacket[9]);
theTans.cfg.datum.dy    = swapFloat(&thePacket[13]);
theTans.cfg.datum.dz    = swapFloat(&thePacket[17]);
}




/*----------------------------------------------------------------------------*/
// Process a TANS Range, Bearing & Distance Packet 0xC8 RNG_BRG_DIST_PKT
/*----------------------------------------------------------------------------*/
void processRngBrgDistPkt(unsigned char * thePacket)
{
theTans.nav.brg        = swapFloat(&thePacket[0]);
theTans.nav.rng        = swapFloat(&thePacket[4]);
theTans.nav.ttg        = swapFloat(&thePacket[12]);
theTans.nav.toWaypoint = swapInt(&thePacket[16]);
}


/*----------------------------------------------------------------------------*/
// Process a TANS Final Distance & ETE Packet 0xC9 FINAL_DIST_ETE_PKT
/*----------------------------------------------------------------------------*/
void processFinalDistPkt(unsigned char * thePacket)
{
theTans.nav.dtg           = swapFloat(&thePacket[0]);
theTans.nav.ete           = swapFloat(&thePacket[4]);
theTans.nav.finalWaypoint = swapInt(&thePacket[8]);
}


/*----------------------------------------------------------------------------*/
// Process a TANS XTE< VTE & HDG error Packet 0xCA XTE_VTE_PKT
/*----------------------------------------------------------------------------*/
void processCrossTrkPkt(unsigned char * thePacket)
{
theTans.nav.xte = swapFloat(&thePacket[0]);
theTans.nav.vte = swapFloat(&thePacket[4]);
theTans.nav.her = swapFloat(&thePacket[8]);
}



/*----------------------------------------------------------------------------*/
// Process a TANS Position Fix LLA, UTM, MGRS Packet 0xCB POS_LLA_UTM_MGRS_PKT
/*----------------------------------------------------------------------------*/
void processMultiCoordPos(unsigned char * thePacket)
{
theTans.gps.coordSys = (unsigned char) thePacket[0];

switch(theTans.gps.coordSys)
	{
	case LLA_POS:
		theTans.gps.pos.lla.lat = swapFloat(&thePacket[1]);
		theTans.gps.pos.lla.lon = swapFloat(&thePacket[5]);
		theTans.gps.pos.lla.alt = swapFloat(&thePacket[9]);
		theTans.gps.timeOfFix   = swapFloat(&thePacket[13]);
		break;

	case UTM_POS:
		theTans.gps.pos.utm.zone[0]  = (unsigned char) thePacket[1];
		theTans.gps.pos.utm.zone[1]  = (unsigned char) thePacket[2];
		theTans.gps.pos.utm.easting  = swapFloat(&thePacket[3]);
		theTans.gps.pos.utm.northing = swapFloat(&thePacket[7]);
		theTans.gps.pos.utm.alt      = swapFloat(&thePacket[11]);
		theTans.gps.timeOfFix        = swapFloat(&thePacket[15]);
		break;

	case MGRS_POS:
		theTans.gps.pos.mgr.zone[0]      = (unsigned char) thePacket[1];
		theTans.gps.pos.mgr.zone[1]      = (unsigned char) thePacket[2];
		theTans.gps.pos.mgr.squareId[0]  = (unsigned char) thePacket[3];
		theTans.gps.pos.mgr.squareId[1]  = (unsigned char) thePacket[4];
		theTans.gps.pos.mgr.easting      = swapFloat(&thePacket[5]);
		theTans.gps.pos.mgr.northing     = swapFloat(&thePacket[9]);
		theTans.gps.pos.mgr.alt          = swapFloat(&thePacket[13]);
		theTans.gps.timeOfFix            = swapFloat(&thePacket[17]);
		break;
	}
}



/*----------------------------------------------------------------------------*/
// Process a TANS Velocity Fix Packet 0xCC VEL_ENU_PKT
/*----------------------------------------------------------------------------*/
void processMultiCoordVel(unsigned char * thePacket)
{
theTans.gps.vel.coordType = (unsigned char) thePacket[0];

switch(theTans.gps.vel.coordType)
	{
	case ENU_VEL:
		theTans.gps.vel.enu.east      = swapFloat(&thePacket[1]);
		theTans.gps.vel.enu.north     = swapFloat(&thePacket[5]);
		theTans.gps.vel.enu.up        = swapFloat(&thePacket[9]);
		theTans.gps.vel.enu.biasRate  = swapFloat(&thePacket[13]);
		theTans.gps.vel.enu.timeOfFix = swapFloat(&thePacket[17]);
		break;

	case CRS_GS_VEL:
		theTans.gps.vel.crs.crs       = swapFloat(&thePacket[1]);    // degrees
		theTans.gps.vel.crs.spd       = swapFloat(&thePacket[5]);    // meters / sec
		theTans.gps.vel.crs.climbRate = swapFloat(&thePacket[9]);    // m/s
		theTans.gps.vel.crs.biasRate  = swapFloat(&thePacket[13]);   // bias rate 
		theTans.gps.vel.crs.timeOfFix = swapFloat(&thePacket[17]);
		break;
	}
}



/*----------------------------------------------------------------------------*/
// Process a TANS Waypoint Packet 0xCD WYPT_LLA_UTM_MGRS_PKT
/*----------------------------------------------------------------------------*/
void processWaypointPkt(unsigned char * thePacket)
{
unsigned char coordType       = (unsigned char) thePacket[0];
int idx                       = swapInt(&thePacket[1]);


// All coordinate types will have these 3 fields in common
theTans.nav.wpt[idx].coordType = coordType;
theTans.nav.wpt[idx].id        = idx;
		
// clear out the old label
memset((void *) theTans.nav.wpt[idx].label, 
       '\0', 
       sizeof(theTans.nav.wpt[idx].label));

// copy in the new label
memcpy((void *) theTans.nav.wpt[idx].label, 
       (void *) &thePacket[3], 
       12);

switch(coordType)
	{
	case LLA_POS:
		theTans.nav.wpt[idx].pos.lla.lat        = swapFloat(&thePacket[15]);
		theTans.nav.wpt[idx].pos.lla.lon        = swapFloat(&thePacket[19]);
		theTans.nav.wpt[idx].pos.lla.alt        = swapFloat(&thePacket[23]);
		theTans.nav.wpt[idx].timeFlag           = swapInt(&thePacket[27]);
		break;

	case UTM_POS:
		theTans.nav.wpt[idx].pos.utm.zone[0]    = (unsigned char) thePacket[15];
		theTans.nav.wpt[idx].pos.utm.zone[1]    = (unsigned char) thePacket[16];
		theTans.nav.wpt[idx].pos.utm.easting    = swapFloat(&thePacket[17]);
		theTans.nav.wpt[idx].pos.utm.northing   = swapFloat(&thePacket[21]);
		theTans.nav.wpt[idx].pos.utm.alt        = swapFloat(&thePacket[25]);
		theTans.nav.wpt[idx].timeFlag           = swapInt(&thePacket[29]);
		break;

	case MGRS_POS:
		theTans.nav.wpt[idx].pos.mgr.zone[0]     = (unsigned char) thePacket[15];
		theTans.nav.wpt[idx].pos.mgr.zone[1]     = (unsigned char) thePacket[16];
		theTans.nav.wpt[idx].pos.mgr.squareId[0] = (unsigned char) thePacket[17];
		theTans.nav.wpt[idx].pos.mgr.squareId[1] = (unsigned char) thePacket[18];
		theTans.nav.wpt[idx].pos.mgr.easting     = swapFloat(&thePacket[19]);
		theTans.nav.wpt[idx].pos.mgr.northing    = swapFloat(&thePacket[23]);
		theTans.nav.wpt[idx].pos.mgr.alt         = swapFloat(&thePacket[27]);
		theTans.nav.wpt[idx].timeFlag            = swapInt(&thePacket[31]);
		break;
	}
}


/*----------------------------------------------------------------------------*/
// Process a Waypoint Switching Mode packet WPT_SW_MODE_PKT 0xA7
/*----------------------------------------------------------------------------*/
void processWptSwModePkt(unsigned char * thePacket)
{
unsigned char idx;

idx                             = (unsigned char) thePacket[0];
theTans.nav.rte[idx].id         = idx;
theTans.nav.rte[idx].switchMode = (unsigned char) thePacket[1];
theTans.nav.rte[idx].swRadius   = swapFloat(&thePacket[2]);
}


/*----------------------------------------------------------------------------*/
// Process a Waypoint List Packet  WPT_LIST_PKT 0xC0
/*----------------------------------------------------------------------------*/
void processRouteListPkt(unsigned char * thePacket)
{
unsigned char idx;
int i;

idx                             = (unsigned char) thePacket[0];
theTans.nav.rte[idx].id         = idx;
for (i=0; i<MAX_LEGS + 1; i+=2)
	{
	theTans.nav.rte[idx].wpts[i]  = swapInt(&thePacket[1+i]);
	}
}


